# Bob build tool
# Copyright (C) 2016  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ..errors import ParseError, BuildError
from ..stringparser import isTrue, IfExpression
from ..tty import WarnOnce, stepAction, INFO, TRACE, WARNING
from ..utils import check_output, joinLines
from .scm import Scm, ScmAudit, ScmStatus, ScmTaint
from shlex import quote
from textwrap import dedent, indent
from xml.etree import ElementTree
import asyncio
import concurrent.futures
import hashlib
import locale
import os, os.path
import re
import schema
import subprocess

class GitScm(Scm):

    SCHEMA = schema.Schema({
        'scm' : 'git',
        'url' : str,
        schema.Optional('dir') : str,
        schema.Optional('if') : schema.Or(str, IfExpression),
        schema.Optional('branch') : str,
        schema.Optional('tag') : str,
        schema.Optional('commit') : str,
        schema.Optional('rev') : str,
        schema.Optional(schema.Regex('^remote-.*')) : str,
        schema.Optional('sslVerify') : bool,
        schema.Optional('singleBranch') : bool,
        schema.Optional('shallow') : schema.Or(int, str)
    })
    REMOTE_PREFIX = "remote-"

    def __init__(self, spec, overrides=[], secureSSL=None):
        super().__init__(spec, overrides)
        self.__url = spec["url"]
        self.__branch = None
        self.__tag = None
        self.__commit = None
        self.__remotes = {}
        if "rev" in spec:
            rev = spec["rev"]
            if rev.startswith("refs/heads/"):
                self.__branch = rev[11:]
            elif rev.startswith("refs/tags/"):
                self.__tag = rev[10:]
            elif len(rev) == 40:
                self.__commit = rev
            else:
                raise ParseError("Invalid rev format: " + rev)
        self.__branch = spec.get("branch", self.__branch)
        self.__tag = spec.get("tag", self.__tag)
        self.__commit = spec.get("commit", self.__commit)
        if self.__commit:
            # validate commit
            if re.match("^[0-9a-f]{40}$", self.__commit) is None:
                raise ParseError("Invalid commit id: " + str(self.__commit))
        elif not self.__branch and not self.__tag:
            # nothing secified at all -> master branch
            self.__branch = "master"
        self.__dir = spec.get("dir", ".")
        # convert remotes into separate dictionary
        for key, val in spec.items():
            if key.startswith(GitScm.REMOTE_PREFIX):
                stripped_key = key[len(GitScm.REMOTE_PREFIX):] # remove prefix
                if stripped_key == "origin":
                    raise ParseError("Invalid remote name: " + stripped_key)
                self.__remotes.update({stripped_key : val})
        self.__sslVerify = spec.get('sslVerify', secureSSL)
        self.__singleBranch = spec.get('singleBranch')
        self.__shallow = spec.get('shallow')

    def getProperties(self):
        properties = super().getProperties()
        properties.update({
            'scm' : 'git',
            'url' : self.__url,
            'branch' : self.__branch,
            'tag' : self.__tag,
            'commit' : self.__commit,
            'dir' : self.__dir,
            'rev' : ( self.__commit if self.__commit else
                (("refs/tags/" + self.__tag) if self.__tag else
                    ("refs/heads/" + self.__branch))
            ),
            'sslVerify' : self.__sslVerify,
            'singleBranch' : self.__singleBranch,
            'shallow' : self.__shallow,
        })
        for key, val in self.__remotes.items():
            properties.update({GitScm.REMOTE_PREFIX+key : val})
        return properties

    async def invoke(self, invoker):
        # make sure the git directory exists
        if not os.path.isdir(invoker.joinPath(self.__dir, ".git")):
            await invoker.checkCommand(["git", "init", self.__dir])

        # Shallow implies singleBranch
        if self.__singleBranch is None:
            singleBranch = self.__shallow is not None
        else:
            singleBranch = self.__singleBranch
        singleBranch = singleBranch and (self.__branch is not None)

        # setup and update remotes
        remotes = { "origin" : self.__url }
        remotes.update(self.__remotes)
        existingRemotes = await invoker.checkOutputCommand(["git", "remote"], cwd=self.__dir)
        for remote in existingRemotes.split("\n"):
            if remote in remotes:
                cfgUrl = remotes[remote]
                realUrl = await invoker.checkOutputCommand(
                    ["git", "ls-remote", "--get-url", remote], cwd=self.__dir)
                if cfgUrl != realUrl:
                    await invoker.checkCommand(["git", "remote", "set-url", remote, cfgUrl], cwd=self.__dir)
                del remotes[remote]

        # add remaining (new) remotes
        for remote,url in remotes.items():
            addCmd = ["git", "remote", "add", remote, url]
            if singleBranch: addCmd += ["-t", self.__branch]
            await invoker.checkCommand(addCmd, cwd=self.__dir)

        # relax security if requested
        if not self.__sslVerify:
            await invoker.checkCommand(["git", "config", "http.sslVerify", "false"], cwd=self.__dir)

        # Calculate refspec that is used internally. For the user a regular
        # refspec is kept in the git config.

        # Base fetch command with shallow support
        fetchCmd = ["git", "fetch", "-p"]
        if isinstance(self.__shallow, int):
            fetchCmd.append("--depth={}".format(self.__shallow))
        elif isinstance(self.__shallow, str):
            fetchCmd.append("--shallow-since={}".format(self.__shallow))
        fetchCmd.append("origin")

        # Calculate appropriate refspec (all/singleBranch/tag)
        if singleBranch:
            fetchCmd += ["+refs/heads/{0}:refs/remotes/origin/{0}".format(self.__branch)]
        else:
            fetchCmd += ["+refs/heads/*:refs/remotes/origin/*"]
        if self.__tag:
            fetchCmd.append("refs/tags/{0}:refs/tags/{0}".format(self.__tag))

        # do the checkout
        if self.__tag or self.__commit:
            await self.__checkoutTag(invoker, fetchCmd)
        else:
            await self.__checkoutBranch(invoker, fetchCmd)

    async def __checkoutTag(self, invoker, fetchCmd):
        # checkout only if HEAD is invalid
        head = await invoker.callCommand(["git", "rev-parse", "--verify", "-q", "HEAD"],
            stdout=False, cwd=self.__dir)
        if head:
            await invoker.checkCommand(fetchCmd, cwd=self.__dir)
            await invoker.checkCommand(["git", "checkout", "-q",
                self.__commit if self.__commit else "tags/"+self.__tag], cwd=self.__dir)

    async def __checkoutBranch(self, invoker, fetchCmd):
        await invoker.checkCommand(fetchCmd, cwd=self.__dir)
        if await invoker.callCommand(["git", "rev-parse", "--verify", "-q", "HEAD"],
                       stdout=False, cwd=self.__dir):
            # checkout only if HEAD is invalid
            await invoker.checkCommand(["git", "checkout", "-b", self.__branch,
                "remotes/origin/"+self.__branch], cwd=self.__dir)
        elif (await invoker.checkOutputCommand(["git", "rev-parse", "--abbrev-ref", "HEAD"], cwd=self.__dir)) == self.__branch:
            # pull only if on original branch
            await invoker.checkCommand(["git", "merge", "--ff-only", "refs/remotes/origin/"+self.__branch], cwd=self.__dir)
        else:
            invoker.warn("Not updating", self.__dir, "because branch was changed manually...")

    def asDigestScript(self):
        """Return forward compatible stable string describing this git module.

        The format is "url rev-spec dir" where rev-spec depends on the given reference.
        """
        if self.__commit:
            return self.__commit + " " + self.__dir
        elif self.__tag:
            return self.__url + " refs/tags/" + self.__tag + " " + self.__dir
        else:
            return self.__url + " refs/heads/" + self.__branch + " " + self.__dir

    def asJenkins(self, workPath, credentials, options):
        scm = ElementTree.Element("scm", attrib={
            "class" : "hudson.plugins.git.GitSCM",
            "plugin" : "git@2.2.7",
        })
        ElementTree.SubElement(scm, "configVersion").text = "2"

        userconfigs =  ElementTree.SubElement(
                ElementTree.SubElement(scm, "userRemoteConfigs"),
                "hudson.plugins.git.UserRemoteConfig")

        url = ElementTree.SubElement(userconfigs,
            "url")
        url.text = self.__url

        if credentials:
            credentialsId = ElementTree.SubElement(userconfigs,
                         "credentialsId")
            credentialsId.text = credentials

        branch = ElementTree.SubElement(
            ElementTree.SubElement(
                ElementTree.SubElement(scm, "branches"),
                "hudson.plugins.git.BranchSpec"),
            "name")
        if self.__commit:
            branch.text = self.__commit
        elif self.__tag:
            branch.text = "refs/tags/" + self.__tag
        else:
            branch.text = "refs/heads/" + self.__branch

        ElementTree.SubElement(scm, "doGenerateSubmoduleConfigurations").text = "false"
        ElementTree.SubElement(scm, "submoduleCfg", attrib={"class" : "list"})

        extensions = ElementTree.SubElement(scm, "extensions")
        ElementTree.SubElement(
            ElementTree.SubElement(extensions,
                "hudson.plugins.git.extensions.impl.RelativeTargetDirectory"),
            "relativeTargetDir").text = os.path.normpath(os.path.join(workPath, self.__dir))
        # remove untracked files and stale branches
        ElementTree.SubElement(extensions,
            "hudson.plugins.git.extensions.impl.CleanCheckout")
        ElementTree.SubElement(extensions,
            "hudson.plugins.git.extensions.impl.PruneStaleBranch")
        # set git clone options
        shallow = options.get("scm.git.shallow")
        timeout = options.get("scm.git.timeout")
        if shallow is not None or timeout is not None:
            co = ElementTree.SubElement(extensions,
                    "hudson.plugins.git.extensions.impl.CloneOption")
            if shallow is not None:
                try:
                    shallow = int(shallow)
                    if shallow < 0: raise ValueError()
                except ValueError:
                    raise BuildError("Invalid 'git.shallow' option: " + str(shallow))
                if shallow > 0:
                    ElementTree.SubElement(co, "shallow").text = "true"
                    ElementTree.SubElement(co, "noTags").text = "false"
                    ElementTree.SubElement(co, "reference").text = ""
                    ElementTree.SubElement(co, "depth").text = str(shallow)
                    ElementTree.SubElement(co, "honorRefspec").text = "false"

            if timeout is not None:
                try:
                    timeout = int(timeout)
                    if timeout < 0: raise ValueError()
                except ValueError:
                    raise BuildError("Invalid 'git.timeout' option: " + str(shallow))
                if timeout > 0:
                    ElementTree.SubElement(co, "timeout").text = str(timeout)

        if isTrue(options.get("scm.ignore-hooks", "0")):
            ElementTree.SubElement(extensions,
                "hudson.plugins.git.extensions.impl.IgnoreNotifyCommit")

        return scm

    def getDirectory(self):
        return self.__dir

    def isDeterministic(self):
        return bool(self.__tag) or bool(self.__commit)

    def hasJenkinsPlugin(self):
        return True

    def callGit(self, workspacePath, *args):
        cmdLine = ['git']
        cmdLine.extend(args)
        cwd = os.path.join(workspacePath, self.__dir)
        try:
            output = subprocess.check_output(cmdLine, cwd=cwd,
                universal_newlines=True, stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError as e:
            raise BuildError("git error:\n Directory: '{}'\n Command: '{}'\n'{}'".format(
                cwd, " ".join(cmdLine), e.output.rstrip()))
        except OSError as e:
            raise BuildError("Error calling git: " + str(e))
        return output.strip()

    def status(self, workspacePath):
        status = ScmStatus()
        try:
            onCorrectBranch = False
            onTag = False
            output = self.callGit(workspacePath, 'ls-remote' ,'--get-url')
            if output != self.__url:
                status.add(ScmTaint.switched,
                    "> URL: configured: '{}', actual: '{}'".format(self.__url, output))

            if self.__commit:
                output = self.callGit(workspacePath, 'rev-parse', 'HEAD')
                if output != self.__commit:
                    status.add(ScmTaint.switched,
                        "> commit: configured: '{}', actual: '{}'".format(self.__commit, output))
            elif self.__tag:
                output = self.callGit(workspacePath, 'tag', '--points-at', 'HEAD').splitlines()
                if self.__tag not in output:
                    actual = ("'" + ", ".join(output) + "'") if output else "not on any tag"
                    status.add(ScmTaint.switched,
                        "> tag: configured: '{}', actual: {}".format(self.__tag, actual))

                # Need to check if the tag still exists. Otherwise the "git
                # log" command at the end will trip.
                try:
                    self.callGit(workspacePath, 'rev-parse', 'tags/'+self.__tag)
                    onTag = True
                except BuildError:
                    pass
            elif self.__branch:
                output = self.callGit(workspacePath, 'rev-parse', '--abbrev-ref', 'HEAD')
                if output != self.__branch:
                    status.add(ScmTaint.switched,
                        "> branch: configured: '{}', actual: '{}'".format(self.__branch, output))
                else:
                    output = self.callGit(workspacePath, 'log', '--oneline',
                        'refs/remotes/origin/'+self.__branch+'..HEAD')
                    if output:
                        status.add(ScmTaint.unpushed_main,
                            joinLines("> unpushed commits on {}:".format(self.__branch),
                                indent(output, '   ')))
                    onCorrectBranch = True

            # Check for modifications wrt. checked out commit
            output = self.callGit(workspacePath, 'status', '--porcelain')
            if output:
                status.add(ScmTaint.modified, joinLines("> modified:",
                    indent(output, '   ')))

            # The following shows all unpushed commits reachable by any ref
            # (local branches, stash, detached HEAD, etc).
            # Exclude HEAD if the configured branch is checked out to not
            # double-count them. Does not mark the SCM as dirty. Exclude the
            # configured tag too if it is checked out. Otherwise the tag would
            # count as unpushed if it is not on a remote branch.
            what = ['--all', '--not', '--remotes']
            if onCorrectBranch: what.append('HEAD')
            if onTag: what.append("tags/"+self.__tag)
            output = self.callGit(workspacePath, 'log', '--oneline', '--decorate',
                *what)
            if output:
                status.add(ScmTaint.unpushed_local,
                    joinLines("> unpushed local commits:", indent(output, '   ')))

        except BuildError as e:
            status.add(ScmTaint.error, e.slogan)

        return status

    def getAuditSpec(self):
        return ("git", self.__dir, {})

    def hasLiveBuildId(self):
        return True

    async def predictLiveBuildId(self, step):
        if self.__commit:
            return bytes.fromhex(self.__commit)

        with stepAction(step, "LS-REMOTE", self.__url, (INFO, TRACE)) as a:
            if self.__tag:
                # Annotated tags are objects themselves. We need the commit object!
                refs = ["refs/tags/" + self.__tag + '^{}', "refs/tags/" + self.__tag]
            else:
                refs = ["refs/heads/" + self.__branch]
            cmdLine = ['git', 'ls-remote', self.__url] + refs
            try:
                stdout = await check_output(cmdLine, stderr=subprocess.DEVNULL,
                    universal_newlines=True)
                output = stdout.strip()
            except subprocess.CalledProcessError as e:
                a.fail("exit {}".format(e.returncode), WARNING)
                return None
            except OSError as e:
                a.fail("error ({})".format(e))
                return None

            # have we found anything at all?
            if not output:
                a.fail("unknown", WARNING)
                return None

            # See if we got one of our intended refs. Git is generating lines with
            # the following format:
            #
            #   <sha1>\t<refname>
            #
            # Put the output into a dict with the refname as key. Be extra careful
            # and strip out lines not matching this pattern.
            output = {
                commitAndRef[1].strip() : bytes.fromhex(commitAndRef[0].strip())
                for commitAndRef
                in (line.split('\t') for line in output.split('\n'))
                if len(commitAndRef) == 2 }
            for ref in refs:
                if ref in output: return output[ref]

            # uhh, should not happen...
            a.fail("unknown", WARNING)
            return None

    def calcLiveBuildId(self, workspacePath):
        if self.__commit:
            return bytes.fromhex(self.__commit)
        else:
            output = self.callGit(workspacePath, 'rev-parse', 'HEAD').strip()
            return bytes.fromhex(output)

    def getLiveBuildIdSpec(self, workspacePath):
        if self.__commit:
            return "=" + self.__commit
        else:
            return "g" + os.path.join(workspacePath, self.__dir)

    @staticmethod
    def processLiveBuildIdSpec(dir):
        try:
            return subprocess.check_output(["git", "rev-parse", "HEAD"],
                cwd=dir, universal_newlines=True).strip()
        except subprocess.CalledProcessError as e:
            raise BuildError("Git audit failed: " + str(e))
        except OSError as e:
            raise BuildError("Error calling git: " + str(e))

class GitAudit(ScmAudit):

    SCHEMA = schema.Schema({
        'type' : 'git',
        'dir' : str,
        'remotes' : { schema.Optional(str) : str },
        'commit' : str,
        'description' : str,
        'dirty' : bool
    })

    async def _scanDir(self, workspace, dir, extra):
        self.__dir = dir
        dir = os.path.join(workspace, dir)
        try:
            remotes = (await check_output(["git", "remote", "-v"],
                cwd=dir, universal_newlines=True)).split("\n")
            remotes = (r[:-8].split("\t") for r in remotes if r.endswith("(fetch)"))
            self.__remotes = { remote:url for (remote,url) in remotes }

            self.__commit = (await check_output(["git", "rev-parse", "HEAD"],
                cwd=dir, universal_newlines=True)).strip()
            self.__description = (await check_output(
                ["git", "describe", "--always", "--dirty=-dirty"],
                cwd=dir, universal_newlines=True)).strip()
            self.__dirty = self.__description.endswith("-dirty")
        except subprocess.CalledProcessError as e:
            raise BuildError("Git audit failed: " + str(e))
        except OSError as e:
            raise BuildError("Error calling git: " + str(e))

    def _load(self, data):
        self.__dir = data["dir"]
        self.__remotes = data["remotes"]
        self.__commit = data["commit"]
        self.__description = data["description"]
        self.__dirty = data["dirty"]

    def dump(self):
        return {
            "type" : "git",
            "dir" : self.__dir,
            "remotes" : self.__remotes,
            "commit" : self.__commit,
            "description" : self.__description,
            "dirty" : self.__dirty,
        }

    def getStatusLine(self):
        return self.__description



# Bob build tool
# Copyright (C) 2016  TechniSat Digital GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys

DEFAULT = 0
SKIPPED = 1
EXECUTED = 2
INFO = 3
WARNING = 4
ERROR = 5
HEADLINE = 6

ALWAYS = -2
IMPORTANT = -1
NORMAL = 0
INFO = 1
DEBUG = 2
TRACE = 3

COLORS2CODE = [ "", "", "32", "34", "33", "31", "32;1" ]
COLORS2TEXT = [ "NOTE", "NOTE", "NOTE", "INFO", "WARN", "ERR ", "====" ]

def colorize(string, color):
    if isinstance(color, int):
        color = COLORS2CODE[color]
    if __useColor and color:
        return "\x1b[" + color + "m" + string + "\x1b[0m"
    else:
        return string

class Unbuffered(object):
    def __init__(self, stream):
        self.stream = stream
    def write(self, data):
        self.stream.write(data)
        self.stream.flush()
    def __getattr__(self, attr):
        return getattr(self.stream, attr)

class Show:
    def __init__(self, slogan, color, message, help, onlyOnce=False):
        self.__slogan = slogan
        self.__color = color
        self.__message = message
        self.__help = help
        self.__triggered = False
        self.__onlyOnce = onlyOnce

    def show(self, location=None):
        if not self.__triggered:
            print(colorize(self.__slogan + ":", self.__color+";1"),
                colorize(((location + ": ") if location else "") + self.__message,
                    self.__color),
                file=sys.stderr)
            if self.__help:
                print(self.__help, file=sys.stderr)
            self.__triggered = self.__onlyOnce

class Info(Show):
    def __init__(self, message, help=None, onlyOnce=False):
        super().__init__("INFO", "34", message, help, onlyOnce)

class InfoOnce(Info):
    def __init__(self, message, help=None):
        super().__init__(message, help, True)

class Warn(Show):
    def __init__(self, message, help=None, onlyOnce=False):
        super().__init__("WARNING", "33", message, help, onlyOnce)

    def warn(self, location=None):
        super().show(location)

class WarnOnce(Warn):
    def __init__(self, message, help=None):
        super().__init__(message, help, True)


###############################################################################

class BaseTUIAction:

    def __init__(self, showDetails):
        self.showDetails = showDetails
        self.ok_kind = EXECUTED
        self.ok_message = "ok"
        self.err_kind = WARNING
        self.err_message = "error"

    def setResult(self, message, kind=EXECUTED, details=""):
        if self.showDetails and details:
            message += " (" + details + ")"
        self.ok_message = message
        self.ok_kind = kind

    def setError(self, message, kind=ERROR, details=""):
        if self.showDetails and details:
            message += " (" + details + ")"
        self.err_message = message
        self.err_kind = kind

    def fail(self, message, kind=ERROR, details=""):
        self.setResult(message, kind, details)
        self.setError(message, kind, details)

class BaseTUI:
    def __init__(self, verbosity):
        self.__verbosity = verbosity

    def getVerbosity(self):
        return self.__verbosity

    def setVerbosity(self, verbosity):
        self.__verbosity = verbosity

    def cleanup(self):
        pass

    def setProgress(self, done, num):
        pass

    def _isVisible(self, severity):
        if isinstance(severity, int):
            return severity <= self.__verbosity
        else:
            low, high = severity
            return (low <= self.__verbosity) and (self.__verbosity <= high)

class DummyTUIAction(BaseTUIAction):
    def __init__(self):
        super().__init__(3)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return False

class SingleTUIAction(BaseTUIAction):

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            kind = self.ok_kind
            message = self.ok_message
        else:
            kind = self.err_kind
            message = self.err_message
        print(colorize(message, kind))
        return False

class SingleTUI(BaseTUI):
    def __init__(self, verbosity):
        super().__init__(verbosity)
        self.__currentPackage = None

    def __setPackage(self, step):
        package = "/".join(step.getPackage().getStack())
        if package != self.__currentPackage:
            self.__currentPackage = package
            print(">>", colorize(self.__currentPackage, HEADLINE))

    def log(self, message, kind, severity):
        if not self._isVisible(severity): return
        print(colorize("** {}".format(message), kind))

    def stepMessage(self, step, action, message, kind, severity):
        if not self._isVisible(severity): return
        self.__setPackage(step)
        print(colorize("   {:10}{}".format(action, message), kind))

    def stepAction(self, step, action, message, severity, details):
        return self.__action(step, action, message, severity, details, True)

    def stepExec(self, step, action, message, severity, details):
        return self.__action(step, action, message, severity, details, False)

    def __action(self, step, action, message, severity, details, ellipsis):
        if not self._isVisible(severity): return DummyTUIAction()
        self.__setPackage(step)
        showDetails = self._isVisible(INFO)
        if showDetails and details:
            details = " " + details
        else:
            details = ""
        if ellipsis:
            print(colorize("   {:10}{}{} .. ".format(action, message, details), EXECUTED), end="")
            return SingleTUIAction(showDetails)
        else:
            print(colorize("   {:10}{}{}".format(action, message, details), EXECUTED))
            return DummyTUIAction()


class ParallelTtyUIAction(BaseTUIAction):
    def __init__(self, tui, job, slot, name, msg, ellipsis, showDetails):
        super().__init__(showDetails)
        self.__tui = tui
        self.__job = job
        self.__slot = slot
        self.__name = name
        self.__msg = msg
        self.__ellipsis = ellipsis
        if not ellipsis: self.setError("")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            kind = ERROR
        else:
            kind = EXECUTED
        msg = "[{:>4}] {}".format(self.__job, colorize(self.__msg, kind))
        if self.__ellipsis:
            if exc_type is None:
                kind = self.ok_kind
                status = self.ok_message
            else:
                kind = self.err_kind
                status = self.err_message
            msg += colorize(status, kind)
        if not self.__ellipsis and exc_type is not None and self.err_message:
            msg = [msg] + [
                "[{:>4}] |{}| {}".format(self.__job, colorize(self.__name, self.err_kind), l)
                for l in self.err_message.split("\n")]
        self.__tui._putResult(self.__slot, msg)
        return False

class ParallelTtyUI(BaseTUI):
    def __init__(self, verbosity, maxJobs):
        super().__init__(verbosity)
        self.__index = 1
        self.__maxJobs = maxJobs
        self.__jobs = {}
        self.__slots = [None] * maxJobs
        self.__tasksDone = 0
        self.__tasksNum = 1

        # disable cursor
        print("\x1b[?25l")

        # disable echo
        try:
            import termios
            fd = sys.stdin.fileno()
            self.__oldTcAttr = termios.tcgetattr(fd)
            new = termios.tcgetattr(fd)
            new[3] = new[3] & ~termios.ECHO
            termios.tcsetattr(fd, termios.TCSADRAIN, new)
        except ImportError:
            pass

    def __nextJob(self):
        ret = self.__index
        self.__index += 1
        return ret

    def __putLineCont(self, line):
        print("\r" + "\x1b[2K", line, "\x1b[K", sep="")

    def __putLine(self, line):
        self.__putLineCont(line)
        self.__putFooter()

    def __putFooter(self):
        print("\r\x1b[?7l\x1b[2K====== {}/{} jobs running, {}% ({}/{} tasks) done "
                .format(len(self.__jobs), self.__maxJobs,
                        self.__tasksDone*100//self.__tasksNum,
                        self.__tasksDone, self.__tasksNum),
              end="")
        i = 0
        while i < self.__maxJobs:
            num = self.__slots[i]
            if num is not None:
                print("\n\x1b[2K {:>4}  {}".format(num, self.__jobs[num]), end='')
            else:
                print("\n\x1b[2K ****  <idle>", end='')
            i += 1
        print("\x1b[{}A".format(i), "\x1b[?7h\r", sep='', end='')

    def _putResult(self, slot, msg):
        job = self.__slots[slot]
        self.__slots[slot] = None
        del self.__jobs[job]
        if msg:
            if isinstance(msg, list):
                for l in msg: self.__putLineCont(l)
                self.__putFooter()
            else:
                self.__putLine(msg)
        else:
            self.__putFooter()

    def log(self, message, kind, severity):
        if not self._isVisible(severity): return
        print(colorize("[****] {}".format(message), kind))

    def stepMessage(self, step, action, message, kind, severity):
        if not self._isVisible(severity): return
        self.__putLine("[    ] {}".format(colorize(
            "{:10}{} - {}".format(action, step.getPackage().getName(), message), kind)))

    def stepAction(self, step, action, message, severity, details):
        return self.__action(step, action, message, severity, details, True)

    def stepExec(self, step, action, message, severity, details):
        return self.__action(step, action, message, severity, details, False)

    def __action(self, step, action, message, severity, details, ellipsis):
        if not self._isVisible(severity): return DummyTUIAction()
        showDetails = self._isVisible(INFO)
        if showDetails and details:
            details = " " + details
        else:
            details = ""
        if ellipsis:
            details += " .. "

        job = self.__nextJob()
        slot = 0
        while self.__slots[slot] is not None: slot += 1
        name = step.getPackage().getName()
        self.__slots[slot] = job
        self.__jobs[job] = colorize("{:10}{} - {}".format(action, name, message), EXECUTED)
        self.__putFooter()
        msg = "{:10}{} - {}{}".format(action, name, message, details)
        return ParallelTtyUIAction(self, job, slot, name, msg, ellipsis, showDetails)

    def cleanup(self):
        self.__putFooter()
        for i in range(max(len(self.__jobs), self.__maxJobs)+1):
            print()
        print("\x1b[?25h")
        try:
            import termios
            termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, self.__oldTcAttr)
        except ImportError:
            pass

    def setProgress(self, done, num):
        self.__tasksDone = done
        self.__tasksNum = num


class ParallelDumbUIAction(BaseTUIAction):
    def __init__(self, tui, job, name, msg, ellipsis, showDetails):
        super().__init__(showDetails)
        self.__tui = tui
        self.__job = job
        self.__name = name
        self.__msg = msg
        self.__ellipsis = ellipsis
        if not ellipsis: self.setError("")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        kind = EXECUTED if exc_type is None else ERROR
        msg = self.__msg
        stderr = None
        if self.__ellipsis:
            if exc_type is None:
                kind = self.ok_kind
                status = self.ok_message
            else:
                kind = self.err_kind
                status = self.err_message
            msg += status
        elif exc_type is not None and self.err_message:
            kind = self.err_kind
            stderr = self.err_message
        self.__tui._print(self.__job, msg, kind, "End")
        if stderr:
            # Print error messages on stderr when being on a dumb output. It is
            # probably redirected by some other script or an analyzed IDE (thin
            # "bob project").
            print(stderr, file=sys.stderr)
        return False

class ParallelDumbUI(BaseTUI):
    def __init__(self, verbosity):
        super().__init__(verbosity)
        self.__index = 1

    def __nextJob(self):
        ret = self.__index
        self.__index += 1
        return ret

    def _print(self, job, msg, kind, stage=""):
        level = COLORS2TEXT[kind & 7]
        print("[{:<5} {:>4}] {}: {}".format(stage, job, level, colorize(msg, kind)))

    def log(self, message, kind, severity):
        if not self._isVisible(severity): return
        self._print("****", message, kind, "*****")

    def stepMessage(self, step, action, message, kind, severity):
        if not self._isVisible(severity): return
        self._print("", "{:10}{} - {}".format(action,
            step.getPackage().getName(), message), kind)

    def stepAction(self, step, action, message, severity, details):
        return self.__action(step, action, message, severity, details, True)

    def stepExec(self, step, action, message, severity, details):
        return self.__action(step, action, message, severity, details, False)

    def __action(self, step, action, message, severity, details, ellipsis):
        if not self._isVisible(severity): return DummyTUIAction()
        showDetails = self._isVisible(INFO)
        if showDetails and details:
            details = " " + details
        else:
            details = ""
        if ellipsis:
            details += ": "

        job = self.__nextJob()
        name = step.getPackage().getName()
        self._print(job, "{:10}{} - {}".format(action, name, message), EXECUTED, "Start")
        msg = "{:10}{} - {}{}".format(action, name, message, details)
        return ParallelDumbUIAction(self, job, name, msg, ellipsis, showDetails)

def log(message, kind, severity=-2):
    __tui.log(message, kind, severity)

def stepMessage(step, action, message, kind, severity=-2):
    __tui.stepMessage(step, action, message, kind, severity)

def stepAction(step, action, message, severity=-2, details=""):
    return __tui.stepAction(step, action, message, severity, details)

def stepExec(step, action, message, severity=-2, details=""):
    return __tui.stepExec(step, action, message, severity, details)

def setVerbosity(verbosity):
    verbosity = max(ALWAYS, min(TRACE, verbosity))
    __tui.setVerbosity(verbosity)

def setProgress(done, num):
    __tui.setProgress(done, num)

def setTui(maxJobs):
    global __tui
    __tui.cleanup()
    if maxJobs <= 1:
        __tui = SingleTUI(__tui.getVerbosity())
    elif __onTTY:
        __tui = ParallelTtyUI(__tui.getVerbosity(), maxJobs)
    else:
        __tui = ParallelDumbUI(__tui.getVerbosity())

def cleanup():
    __tui.cleanup()
    if __onTTY and sys.platform == "win32":
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), __origMode.value)

def ttyReinit():
    """Re-initialize the console settings.

    Work around a MSYS2 odity where the executable unconditionally resets the
    ENABLE_VIRTUAL_TERMINAL_PROCESSING flag even if it was already set when the
    process was started.
    """
    if __onTTY and sys.platform == "win32":
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), __origMode.value | 4)

# module initialization

__onTTY = (sys.stdout.isatty() and sys.stderr.isatty())
__useColor = False
__tui = SingleTUI(NORMAL)

if __onTTY and sys.platform == "win32":
    # Try to set ENABLE_VIRTUAL_TERMINAL_PROCESSING flag. Enables vt100 color
    # codes on Windows 10 console. If this fails we inhibit color code usage
    # because it will clutter the output.
    import ctypes
    import ctypes.wintypes
    __origMode = ctypes.wintypes.DWORD()
    kernel32 = ctypes.windll.kernel32
    kernel32.GetConsoleMode(kernel32.GetStdHandle(-11), ctypes.byref(__origMode))
    if not kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), __origMode.value | 4):
        __onTTY = False

def setColorMode(mode):
    global __useColor
    if mode == 'never':
        __useColor = False
    elif mode == 'always':
        __useColor = True
    elif mode == 'auto':
        __useColor = __onTTY

# auto is the default
setColorMode('auto')



# Sample plugin showing how to add a plugin to some Jenkins jobs
#
# This plugin adds a cobertura coverage publisher post-build-action to all jobs
# that contain at least one package that ends with 'unittests'.

from xml.etree import ElementTree

PLUGIN = ElementTree.fromstring("""
    <hudson.plugins.cobertura.CoberturaPublisher plugin="cobertura@1.9.8">
      <coberturaReportFile>**/coverage.xml</coberturaReportFile>
      <onlyStable>false</onlyStable>
      <failUnhealthy>false</failUnhealthy>
      <failUnstable>false</failUnstable>
      <autoUpdateHealth>false</autoUpdateHealth>
      <autoUpdateStability>false</autoUpdateStability>
      <zoomCoverageChart>false</zoomCoverageChart>
      <maxNumberOfBuilds>0</maxNumberOfBuilds>
      <failNoReports>true</failNoReports>
      <healthyTarget>
        <targets class="enum-map" enum-type="hudson.plugins.cobertura.targets.CoverageMetric">
          <entry>
            <hudson.plugins.cobertura.targets.CoverageMetric>METHOD</hudson.plugins.cobertura.targets.CoverageMetric>
            <int>8000000</int>
          </entry>
          <entry>
            <hudson.plugins.cobertura.targets.CoverageMetric>LINE</hudson.plugins.cobertura.targets.CoverageMetric>
            <int>8000000</int>
          </entry>
          <entry>
            <hudson.plugins.cobertura.targets.CoverageMetric>CONDITIONAL</hudson.plugins.cobertura.targets.CoverageMetric>
            <int>7000000</int>
          </entry>
        </targets>
      </healthyTarget>
      <unhealthyTarget>
        <targets class="enum-map" enum-type="hudson.plugins.cobertura.targets.CoverageMetric">
          <entry>
            <hudson.plugins.cobertura.targets.CoverageMetric>METHOD</hudson.plugins.cobertura.targets.CoverageMetric>
            <int>0</int>
          </entry>
          <entry>
            <hudson.plugins.cobertura.targets.CoverageMetric>LINE</hudson.plugins.cobertura.targets.CoverageMetric>
            <int>0</int>
          </entry>
          <entry>
            <hudson.plugins.cobertura.targets.CoverageMetric>CONDITIONAL</hudson.plugins.cobertura.targets.CoverageMetric>
            <int>0</int>
          </entry>
        </targets>
      </unhealthyTarget>
      <failingTarget>
        <targets class="enum-map" enum-type="hudson.plugins.cobertura.targets.CoverageMetric">
          <entry>
            <hudson.plugins.cobertura.targets.CoverageMetric>METHOD</hudson.plugins.cobertura.targets.CoverageMetric>
            <int>0</int>
          </entry>
          <entry>
            <hudson.plugins.cobertura.targets.CoverageMetric>LINE</hudson.plugins.cobertura.targets.CoverageMetric>
            <int>0</int>
          </entry>
          <entry>
            <hudson.plugins.cobertura.targets.CoverageMetric>CONDITIONAL</hudson.plugins.cobertura.targets.CoverageMetric>
            <int>0</int>
          </entry>
        </targets>
      </failingTarget>
      <sourceEncoding>ASCII</sourceEncoding>
    </hudson.plugins.cobertura.CoberturaPublisher>
""")

def checkCobertura(config, checkoutSteps, buildSteps, packageSteps, **kwargs):
    found = False
    for s in checkoutSteps:
        if s.getPackage().getName().endswith("unittests"): found = True
    for s in buildSteps:
        if s.getPackage().getName().endswith("unittests"): found = True
    for s in packageSteps:
        if s.getPackage().getName().endswith("unittests"): found = True

    if found:
        root = ElementTree.fromstring(config)
        publishers = root.find("publishers")
        if publishers.find("hudson.plugins.cobertura.CoberturaPublisher") is None:
            publishers.append(PLUGIN)
            config = ElementTree.tostring(root, encoding="UTF-8")

    return config

manifest = {
    'apiVersion' : "0.3",
    'hooks' : {
        'jenkinsJobCreate' : checkCobertura,
        'jenkinsJobPostUpdate' : checkCobertura
    }
}



# Bob build tool
# Copyright (C) 2019  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys

# validate Python version
if sys.version_info.major != 3:
    print("Bob requires Python 3")
    sys.exit(1)
elif sys.version_info.minor < 5:
    print("Bob requires at least Python 3.5")
    sys.exit(1)

def getVersion():
    import os, re

    version = ""
    root = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', '..', '..')

    # try to query git
    if os.path.exists(os.path.join(root, ".git")):
        import subprocess
        try:
            version = subprocess.check_output("git describe --tags --dirty".split(" "),
                cwd=root, universal_newlines=True, stderr=subprocess.DEVNULL)
        except (subprocess.CalledProcessError, OSError):
            pass

    if version:
        m = re.match(r"^v(?P<version>[0-9]+(?:\.[0-9]+){2})(?P<rc>-rc[0-9]+)?(?P<dist>-[0-9]+-g[a-f0-9]+)?(?P<dirty>-dirty)?$", version)
        if m is not None:
            # Convert to PEP 440 conforming version number
            version = [ int(i) for i in m.group("version").split(".") ]
            local = []

            if m.group("rc"):
                version.append("rc")
                version.append(int(m.group("rc")[3:]))
            if m.group("dist"):
                # Development versions guess the next cut. This is done by
                # simply incrementing the last version number (might be patch
                # or rc).
                dist,commit = m.group("dist")[1:].split("-")
                version[-1] += 1
                version.append(".dev" + dist)
                local.append(commit)
            if m.group("dirty"):
                local.append("dirty")

            version = [ str(i) for i in version ]
            version = ".".join(version[:3]) + "".join(version[3:])
            if local:
                version += "+" + ".".join(local)
        else:
            import sys
            print("Warning: inferred version of Bob does not match schema:",
                version, file=sys.stderr)
            version = ""

    if not version:
        # Last fallback. See PEP 440 and adjust accordingly.
        version = "0.17+unknown"

    return version

try:
    BOB_VERSION = getVersion()
except KeyboardInterrupt:
    sys.exit(1)




# Bob build tool
# Copyright (C) 2017  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ..errors import ParseError
from ..utils import joinLines
from abc import ABCMeta, abstractmethod
from enum import Enum
from shlex import quote
import fnmatch
import re

class ScmOverride:
    def __init__(self, override):
        self.__match = override.get("match", {})
        self.__del = override.get("del", [])
        self.__set = override.get("set", {})
        self.__if = override.get("if", None)
        self.__replaceRaw = override.get("replace", {})
        self.__init()

    def __init(self):
        try:
            self.__replace = { key : (re.compile(subst["pattern"]), subst["replacement"])
                for (key, subst) in self.__replaceRaw.items() }
        except re.error as e:
            raise ParseError("Invalid scmOverrides replace pattern: '{}': {}"
                .format(e.pattern, str(e)))

    def __getstate__(self):
        return (self.__match, self.__del, self.__set, self.__replaceRaw, self.__if)

    def __setstate__(self, s):
        (self.__match, self.__del, self.__set, self.__replaceRaw, self.__if) = s
        self.__init()

    def __doesMatch(self, scm, env):
        if self.__if is not None and not env.evaluate(self.__if, "scmOverride::if"): return False
        for (key, value) in self.__match.items():
            if key not in scm: return False
            if type(scm[key]) != type(value): return False
            if isinstance(value, str):
                value = env.substitute(value, "svmOverride::match")
                if not fnmatch.fnmatchcase(scm[key], value): return False
            else:
                if scm[key] != value: return False
        return True

    def __hash__(self):
        return hash((frozenset(self.__match.items()), frozenset(self.__del),
            frozenset(self.__set.items()), frozenset(self.__replace.items())))

    def __eq__(self, other):
        return ((self.__match, self.__del, self.__set, self.__replace) ==
            (other.__match, other.__del, other.__set, other.__replace))

    def __applyEnv(self, env):
        rm = [ env.substitute(d, "svmOverrides::del") for d in self.__del ]
        set = {
            k : env.substitute(v, "svmOverrides::set: "+k) if isinstance(v, str) else v
            for (k,v) in self.__set.items()
        }
        return rm, set

    def mangle(self, scm, env):
        ret = False
        if self.__doesMatch(scm, env):
            rm, set = self.__applyEnv(env)

            ret = True
            scm = scm.copy()
            for d in rm:
                if d in scm: del scm[d]
            scm.update(set)
            for (key, (pat, repl)) in self.__replace.items():
                if key in scm:
                    scm[key] = re.sub(pat, repl, scm[key])
        return ret, scm

    def __str__(self):
        import yaml
        spec = {}
        if self.__match: spec['match'] = self.__match
        if self.__del: spec['del'] = self.__del
        if self.__set: spec['set'] = self.__set
        if self.__replaceRaw: spec['replace'] = self.__replaceRaw
        return yaml.dump(spec, default_flow_style=False).rstrip()


class ScmTaint(Enum):
    """
    The taint flags are single letter flags that indicate certain states of the
    SCM.

    Their meaning is as follows:
     - attic - Recipe changed. Will be moved to attic.
     - collides - New checkout but obstructed by existing file.
     - error - Something went really wrong when getting status.
     - modified -  The SCM has been modified wrt. checked out commit.
     - new - New checkout.
     - overridden - A scmOverrides entry applies.
     - switched - The SCM branch/tag/commit was changed by the user.
     - unknown - Not enough information to get further status.
     - unpushed_main - Configured branch with commits not in remote.
     - unpushed_local - Some local branch with unpushed commits exists.
    """
    attic = 'A'
    collides = 'C'
    error = 'E'
    modified = 'M'
    new = 'N'
    overridden = 'O'
    switched = 'S'
    unknown = '?'
    unpushed_main = 'U'
    unpushed_local = 'u'

class ScmStatus:
    """"
    Describes an SCM status wrt. recipe.

    The important status is stored as a set of ScmTaint flags. Additionally the
    'description' field holds any output from the SCM tool that is interesting
    to the user to judge the SCM status. This is only shown in verbose output
    mode.
    """

    def __init__(self, flag=None, description=""):
        self.__flags = {}
        if flag is not None:
            self.__flags[flag] = description

    def __str__(self):
        return "".join(sorted(f.value for f in self.flags))

    @property
    def clean(self):
        """
        Is SCM branch/tag/commit the same as specified in the recipe and no
        local changes?
        """
        return not self.dirty

    @property
    def dirty(self):
        """
        Is SCM is dirty?

        Could be: errors, modified files or switched to another
        branch/tag/commit/repo.  Unpushed commits on the configured branch also
        count as dirty because they are locally commited changes that are not
        visible upstream. On the other hand unpushed changes on unrelated
        branches (unpushed_local) do not count.
        """
        return bool(self.flags & {ScmTaint.modified, ScmTaint.error,
            ScmTaint.switched, ScmTaint.unpushed_main})

    @property
    def error(self):
        """
        Check if SCM is in an error state.

        Set if the SCM command returned a error code or something unexpected
        happened while gathering the status.
        """
        return ScmTaint.error in self.flags

    @property
    def expendable(self):
        """
        Could the SCM be deleted without loosing user data?

        This is more strict than 'dirty' because it includes unrelated local
        branches that the user might have created.
        """
        return not self.dirty and self.flags.isdisjoint(
            {ScmTaint.unpushed_local, ScmTaint.unknown})

    @property
    def flags(self):
        return frozenset(self.__flags.keys())

    def description(self, subset=None):
        if subset:
            flags = {
                flag : description for flag,description in self.__flags.items()
                if flag in subset
            }
        else:
            flags = self.__flags

        # join active descriptions sorted by flag value
        return joinLines(*(d for f,d in
            sorted(flags.items(), key=lambda x: x[0].value)))

    def add(self, flag, description=""):
        if flag in self.__flags:
            self.__flags[flag] = joinLines(self.__flags[flag], description)
        else:
            self.__flags[flag] = description

    def merge(self, other):
        for flag,description in other.__flags.items():
            self.add(flag, description)


class Scm(metaclass=ABCMeta):
    def __init__(self, spec, overrides):
        # Recipe foobar, checkoutSCM dir:., url:asdf
        self.__source = spec.get("__source", "<unknown>") + " in checkoutSCM: dir:" + \
            spec.get("dir", ".") + ", url:" + spec.get("url", "?")
        self.__recipe = spec["recipe"]
        self.__overrides = overrides

    def getSource(self):
        return self.__source

    def getProperties(self):
        return {
            "__source" : self.__source,
            "recipe" : self.__recipe,
        }

    @abstractmethod
    async def invoke(self, invoker):
        """Execute the SCM checkout with the passed invoker instance.

        Everything must be done with the passed invoker instance. It will be
        configured for the right workspace and will do the logging, error
        handling and so on...
        """

    @abstractmethod
    def asDigestScript(self):
        """Return forward compatible stable string describing this SCM.

        The string should represent what the SCM checks out. This is different
        from the actual actions that are returned by asScript() or asJenkins()
        which might evolve in future versions. The returned string is used to
        compute the various IDs and to detect changes to the SDM.
        """
        return ""

    def asJenkins(self, workPath, credentials, options):
        """Return Jenkins xml.etree.ElementTree fragment that does the checkout.

        This is only called if hasJenkinsPlugin() returns True. In this case
        asScript() is not used on Jenkins.
        """
        return None

    def hasJenkinsPlugin(self):
        """Does this SCM use a Jenins plugin?"""
        return False

    @abstractmethod
    def getDirectory(self):
        """Return relative directory that this SCM owns in the workspace."""
        return ""

    @abstractmethod
    def isDeterministic(self):
        """Return whether the SCM is deterministic."""
        return False

    def status(self, workspacePath):
        """Get SCM work-space status.

        The purpose of this method is to return the status of the given
        directory in the work-space. The returned value is used for 'bob
        status' and to implement --clean-checkout. Shall return a ScmStatus()
        object.

        This method is called when building with --clean-checkout. If the
        returned ScmStatus objects 'error' or 'dirty' properties are True then
        the SCM is moved to the attic, while clean directories are not.
        """

        return ScmStatus()

    def getActiveOverrides(self):
        """Return list of ScmOverride objects that matched this SCM."""
        return self.__overrides

    def getAuditSpec(self):
        """Return spec for audit trail generation.

        Must return a tuple of three elements. The first element is a string that
        is used to find the right Audit class (see bob.audit.Artifact.SCMS).
        The second element is a relative directory in the workspace that must
        be audited. The third element is a dict with additional meta information
        that is passed to the audit scanner.

        If the SCM does not support audit trail generation then None shall be
        returned.
        """
        return None

    def hasLiveBuildId(self):
        """Check if live build-ids are supported."""
        return False

    async def predictLiveBuildId(self, step):
        """Query server to predict live build-id."""
        return None

    def calcLiveBuildId(self, workspacePath):
        """Calculate live build-id from workspace."""
        return None

    def getLiveBuildIdSpec(self, workspacePath):
        """Generate spec lines for bob-hash-engine."""
        return None

class ScmAudit(metaclass=ABCMeta):
    @classmethod
    async def fromDir(cls, workspace, dir, extra):
        """Create SCM audit record by scanning a directory"""
        scm = cls()
        await scm._scanDir(workspace, dir, extra)
        return scm

    @classmethod
    def fromData(cls, data):
        """Restore SCM audit from serialized record"""
        scm = cls()
        scm._load(data)
        return scm

    @abstractmethod
    async def _scanDir(self, workspace, dir):
        """Scan directory for SCM"""
        pass

    @abstractmethod
    def _load(self, data):
        """Load from persisted record"""
        pass

    @abstractmethod
    def dump(self):
        """Serialize state into an ElementTree.Element"""
        pass

    def getStatusLine(self):
        return "unknown"



# Bob Build Tool
# Copyright (C) 2019  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ..errors import BuildError
from ..invoker import Invoker, InvocationMode
from ..languages import StepSpec
from ..utils import EventLoopWrapper
import argparse
import sys


def doInvoke(argv, bobRoot):
    parser = argparse.ArgumentParser(prog="bob _invoke",
        description="Invoke a single step.")
    parser.add_argument('spec', help="The step spec file")
    parser.add_argument('mode', default='run', choices=['run', 'shell', 'fingerprint'], nargs='?',
        help="Invocation mode")

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--clean', '-c', action='store_true', default=False,
        help="Clean workspace before execution")
    group.add_argument('--incremental', '-i', action='store_false', dest='clean',
        help="Do not clean workspace before execution")

    parser.add_argument('--keep-sandbox', '-k', action='store_true',
        help="Keep sandbox after execution")
    parser.add_argument('-E', dest="preserve_env", default=False, action='store_true',
        help="Preserve whole environment")
    parser.add_argument('-n', '--no-logfiles', default=False, action='store_true',
        help="Disable log file generation.")
    parser.add_argument('-q', '--quiet', default=0, action='count',
        help="Decrease verbosity (may be specified multiple times)")
    parser.add_argument('-v', '--verbose', default=0, action='count',
        help="Increase verbosity (may be specified multiple times)")

    args = parser.parse_args(argv)
    verbosity = min(3, max(0, 1 + args.verbose - args.quiet)) # [0..4], default: 1

    try:
        with open(args.spec) as f:
            spec = StepSpec.fromFile(f)
    except OSError as e:
        raise BuildError("Error reading spec: " + str(e))

    # Let's do it...
    with EventLoopWrapper() as loop:
        if args.mode == 'shell':
            invoker = Invoker(spec, args.preserve_env, True, True, True, False, False)
            ret = loop.run_until_complete(invoker.executeStep(InvocationMode.SHELL,
                args.clean, args.keep_sandbox))
        elif args.mode == 'run':
            invoker = Invoker(spec, args.preserve_env, args.no_logfiles,
                verbosity >= 2, verbosity >= 1, verbosity >= 3, False)
            ret = loop.run_until_complete(invoker.executeStep(InvocationMode.CALL,
                args.clean, args.keep_sandbox))
        elif args.mode == 'fingerprint':
            invoker = Invoker(spec, args.preserve_env, True, True, True, verbosity >= 3, False)
            (ret, stdout, stderr) = loop.run_until_complete(invoker.executeFingerprint(args.keep_sandbox))
            if ret == 0:
                sys.stdout.buffer.write(stdout)
            else:
                sys.stderr.buffer.write(stderr)
        else:
            assert False, "not reached"

    # Convert signals to error codes like bash does
    if ret < 0:
        ret = 128 - ret
    return ret



# Bob build tool
# Copyright (C) 2016-2020 The BobBuildTool Contributors
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Binary archive handling module.

Every backend is expected to implement the following behaviour:

Downloads will almost never throw a fatal error. It should make no difference
if an artifact was not found or could not be read. Bob should move on to the
next backend or build the package. The only exception is when an artifact could
be found and there are errors during extraction (e.g. some sanity checks fail,
corrupted archive, no space left, ...).

Uploads are supposed to throw a BuildError when something goes wrong unless the
'nofail' option is set. If the artifact is already in the artifact repository
it must not be overwritten. The backend should make sure that even on
concurrent uploads the artifact must appear atomically for unrelated readers.
"""

from . import BOB_VERSION
from .errors import BuildError
from .tty import stepAction, stepMessage, \
    SKIPPED, EXECUTED, WARNING, INFO, TRACE, ERROR, IMPORTANT
from .utils import asHexStr, removePath, isWindows
from shlex import quote
from tempfile import mkstemp, NamedTemporaryFile, TemporaryFile
import argparse
import asyncio
import concurrent.futures
import concurrent.futures.process
import gzip
import hashlib
import http.client
import os
import os.path
import signal
import ssl
import subprocess
import tarfile
import textwrap
import urllib.parse

ARCHIVE_GENERATION = '-1'
ARTIFACT_SUFFIX = ".tgz"
BUILDID_SUFFIX = ".buildid"
FINGERPRINT_SUFFIX = ".fprnt"

def buildIdToName(bid):
    return asHexStr(bid) + ARCHIVE_GENERATION

def readFileOrHandle(name, fileobj):
    if fileobj is not None:
        return fileobj.read()
    with open(name, "rb") as f:
        return f.read()

def writeFileOrHandle(name, fileobj, content):
    if fileobj is not None:
        fileobj.write(content)
        return
    with open(name, "wb") as f:
        f.write(content)


class DummyArchive:
    """Archive that does nothing"""

    def wantDownload(self, enable):
        pass

    def wantUpload(self, enable):
        pass

    def canDownloadLocal(self):
        return False

    def canUploadLocal(self):
        return False

    def canDownloadJenkins(self):
        return False

    def canUploadJenkins(self):
        return False

    async def uploadPackage(self, step, buildId, audit, content):
        pass

    async def downloadPackage(self, step, buildId, audit, content):
        return False

    def upload(self, step, buildIdFile, tgzFile):
        return ""

    def download(self, step, buildIdFile, tgzFile):
        return ""

    async def uploadLocalLiveBuildId(self, step, liveBuildId, buildId):
        pass

    async def downloadLocalLiveBuildId(self, step, liveBuildId):
        return None

    def uploadJenkinsLiveBuildId(self, step, liveBuildId, buildId, isWin):
        return ""

    async def uploadLocalFingerprint(self, step, key, fingerprint):
        pass

    async def downloadLocalFingerprint(self, step, key):
        return None

    def uploadJenkinsFingerprint(self, step, keyFile, fingerprintFile):
        return ""

class ArtifactNotFoundError(Exception):
    pass

class ArtifactExistsError(Exception):
    pass

class ArtifactDownloadError(Exception):
    def __init__(self, reason):
        self.reason = reason

class ArtifactUploadError(Exception):
    def __init__(self, reason):
        self.reason = reason

class BaseArchive:
    def __init__(self, spec):
        flags = spec.get("flags", ["upload", "download"])
        self.__useDownload = "download" in flags
        self.__useUpload = "upload" in flags
        self.__ignoreErrors = "nofail" in flags
        self.__useLocal = "nolocal" not in flags
        self.__useJenkins = "nojenkins" not in flags
        self.__wantDownload = False
        self.__wantUpload = False

    def _ignoreErrors(self):
        return self.__ignoreErrors

    def wantDownload(self, enable):
        self.__wantDownload = enable

    def wantUpload(self, enable):
        self.__wantUpload = enable

    def canDownloadLocal(self):
        return self.__wantDownload and self.__useDownload and self.__useLocal

    def canUploadLocal(self):
        return self.__wantUpload and self.__useUpload and self.__useLocal

    def canDownloadJenkins(self):
        return self.__wantDownload and self.__useDownload and self.__useJenkins

    def canUploadJenkins(self):
        return self.__wantUpload and self.__useUpload and self.__useJenkins

    def __extractPackage(self, tar, audit, content):
        if tar.pax_headers.get('bob-archive-vsn', "0") != "1":
            raise BuildError("Unsupported binary artifact")

        f = tar.next()
        while f is not None:
            if f.name.startswith("content/"):
                if f.islnk():
                    if not f.linkname.startswith("content/"):
                        raise BuildError("invalid hard link in archive: '{}' -> '{}'"
                                            .format(f.name, f.linkname))
                    f.linkname = f.linkname[8:]
                f.name = f.name[8:]
                try:
                    tar.extract(f, content)
                except UnicodeError:
                    raise BuildError("File name encoding error while extracting '{}'".format(f.name),
                                     help="Your locale(7) probably does not (fully) support unicode.")
            elif f.name == "meta/audit.json.gz":
                f.name = audit
                tar.extract(f)
            elif f.name == "content" or f.name == "meta":
                pass
            else:
                raise BuildError("Binary artifact contained unknown file: " + f.name)
            f = tar.next()

    def _openDownloadFile(self, buildId, suffix):
        raise ArtifactNotFoundError()

    async def downloadPackage(self, step, buildId, audit, content):
        if not self.canDownloadLocal():
            return False

        loop = asyncio.get_event_loop()
        suffix = ARTIFACT_SUFFIX
        details = " from {}".format(self._remoteName(buildId, suffix))
        with stepAction(step, "DOWNLOAD", content, details=details) as a:
            try:
                ret, msg, kind = await loop.run_in_executor(None, BaseArchive._downloadPackage,
                    self, buildId, suffix, audit, content)
                if not ret: a.fail(msg, kind)
                return ret
            except (concurrent.futures.CancelledError, concurrent.futures.process.BrokenProcessPool):
                raise BuildError("Download of package interrupted.")

    def _downloadPackage(self, buildId, suffix, audit, content):
        # restore signals to default so that Ctrl+C kills us
        signal.signal(signal.SIGINT, signal.SIG_DFL)

        try:
            with self._openDownloadFile(buildId, suffix) as (name, fileobj):
                with tarfile.open(name, "r|*", fileobj=fileobj, errorlevel=1) as tar:
                    removePath(audit)
                    removePath(content)
                    os.makedirs(content)
                    self.__extractPackage(tar, audit, content)
            return (True, None, None)
        except ArtifactNotFoundError:
            return (False, "not found", WARNING)
        except ArtifactDownloadError as e:
            return (False, e.reason, WARNING)
        except BuildError as e:
            raise
        except OSError as e:
            raise BuildError("Cannot download artifact: " + str(e))
        except tarfile.TarError as e:
            raise BuildError("Error extracting binary artifact: " + str(e))

    async def downloadLocalLiveBuildId(self, step, liveBuildId):
        if not self.canDownloadLocal():
            return None

        loop = asyncio.get_event_loop()
        with stepAction(step, "MAP-SRC", self._remoteName(liveBuildId, BUILDID_SUFFIX), (INFO,TRACE)) as a:
            try:
                ret, msg, kind = await loop.run_in_executor(None,
                    BaseArchive._downloadLocalFile, self, liveBuildId, BUILDID_SUFFIX)
                if ret is None: a.fail(msg, kind)
                return ret
            except (concurrent.futures.CancelledError, concurrent.futures.process.BrokenProcessPool):
                raise BuildError("Download of build-id interrupted.")

    def _downloadLocalFile(self, key, suffix):
        # restore signals to default so that Ctrl+C kills us
        signal.signal(signal.SIGINT, signal.SIG_DFL)

        try:
            with self._openDownloadFile(key, suffix) as (name, fileobj):
                ret = readFileOrHandle(name, fileobj)
            return (ret, None, None)
        except ArtifactNotFoundError:
            return (None, "not found", WARNING)
        except ArtifactDownloadError as e:
            return (None, e.reason, WARNING)
        except BuildError as e:
            raise
        except OSError as e:
            raise BuildError("Cannot download file: " + str(e))

    def _openUploadFile(self, buildId, suffix):
        raise ArtifactUploadError("not implemented")

    async def uploadPackage(self, step, buildId, audit, content):
        if not self.canUploadLocal():
            return
        if not audit:
            stepMessage(step, "UPLOAD", "skipped (no audit trail)", SKIPPED,
                IMPORTANT)
            return

        loop = asyncio.get_event_loop()
        suffix = ARTIFACT_SUFFIX
        details = " to {}".format(self._remoteName(buildId, suffix))
        with stepAction(step, "UPLOAD", content, details=details) as a:
            try:
                msg, kind = await loop.run_in_executor(None, BaseArchive._uploadPackage,
                    self, buildId, suffix, audit, content)
                a.setResult(msg, kind)
            except (concurrent.futures.CancelledError, concurrent.futures.process.BrokenProcessPool):
                raise BuildError("Upload of package interrupted.")

    def _uploadPackage(self, buildId, suffix, audit, content):
        # restore signals to default so that Ctrl+C kills us
        signal.signal(signal.SIGINT, signal.SIG_DFL)

        try:
            with self._openUploadFile(buildId, suffix) as (name, fileobj):
                pax = { 'bob-archive-vsn' : "1" }
                with gzip.open(name or fileobj, 'wb', 6) as gzf:
                    with tarfile.open(name, "w", fileobj=gzf,
                                      format=tarfile.PAX_FORMAT, pax_headers=pax) as tar:
                        tar.add(audit, "meta/" + os.path.basename(audit))
                        tar.add(content, arcname="content")
        except ArtifactExistsError:
            return ("skipped ({} exists in archive)".format(content), SKIPPED)
        except (ArtifactUploadError, tarfile.TarError, OSError) as e:
            if self.__ignoreErrors:
                return ("error ("+str(e)+")", ERROR)
            else:
                raise BuildError("Cannot upload artifact: " + str(e))
        return ("ok", EXECUTED)

    async def uploadLocalLiveBuildId(self, step, liveBuildId, buildId):
        if not self.canUploadLocal():
            return

        loop = asyncio.get_event_loop()
        with stepAction(step, "CACHE-BID", self._remoteName(liveBuildId, BUILDID_SUFFIX), (INFO,TRACE)) as a:
            try:
                msg, kind = await loop.run_in_executor(None, BaseArchive._uploadLocalFile, self, liveBuildId, BUILDID_SUFFIX, buildId)
                a.setResult(msg, kind)
            except (concurrent.futures.CancelledError, concurrent.futures.process.BrokenProcessPool):
                raise BuildError("Upload of build-id interrupted.")

    def _uploadLocalFile(self, key, suffix, content):
        # restore signals to default so that Ctrl+C kills us
        signal.signal(signal.SIGINT, signal.SIG_DFL)

        try:
            with self._openUploadFile(key, suffix) as (name, fileobj):
                writeFileOrHandle(name, fileobj, content)
        except ArtifactExistsError:
            return ("skipped (exists in archive)", SKIPPED)
        except (ArtifactUploadError, OSError) as e:
            if self.__ignoreErrors:
                return ("error ("+str(e)+")", ERROR)
            else:
                raise BuildError("Cannot upload file: " + str(e))
        return ("ok", EXECUTED)

    async def uploadLocalFingerprint(self, step, key, fingerprint):
        if not self.canUploadLocal():
            return

        loop = asyncio.get_event_loop()
        with stepAction(step, "CACHE-FPR", self._remoteName(key, FINGERPRINT_SUFFIX)) as a:
            try:
                msg, kind = await loop.run_in_executor(None, BaseArchive._uploadLocalFile, self, key, FINGERPRINT_SUFFIX, fingerprint)
                a.setResult(msg, kind)
            except (concurrent.futures.CancelledError, concurrent.futures.process.BrokenProcessPool):
                raise BuildError("Upload of build-id interrupted.")

    async def downloadLocalFingerprint(self, step, key):
        if not self.canDownloadLocal():
            return None

        loop = asyncio.get_event_loop()
        with stepAction(step, "MAP-FPRNT", self._remoteName(key, FINGERPRINT_SUFFIX)) as a:
            try:
                ret, msg, kind = await loop.run_in_executor(None,
                    BaseArchive._downloadLocalFile, self, key, FINGERPRINT_SUFFIX)
                if ret is None: a.fail(msg, kind)
                return ret
            except (concurrent.futures.CancelledError, concurrent.futures.process.BrokenProcessPool):
                raise BuildError("Download of fingerprint interrupted.")


class LocalArchive(BaseArchive):
    def __init__(self, spec):
        super().__init__(spec)
        self.__basePath = os.path.abspath(spec["path"])
        self.__fileMode = spec.get("fileMode")
        self.__dirMode = spec.get("directoryMode")

    def _getPath(self, buildId, suffix):
        packageResultId = buildIdToName(buildId)
        packageResultPath = os.path.join(self.__basePath, packageResultId[0:2],
                                         packageResultId[2:4])
        packageResultFile = os.path.join(packageResultPath,
                                         packageResultId[4:]) + suffix
        return (packageResultPath, packageResultFile)

    def _remoteName(self, buildId, suffix):
        return self._getPath(buildId, suffix)[1]

    def _openDownloadFile(self, buildId, suffix):
        (packageResultPath, packageResultFile) = self._getPath(buildId, suffix)
        if os.path.isfile(packageResultFile):
            return LocalArchiveDownloader(packageResultFile)
        else:
            raise ArtifactNotFoundError()

    def _openUploadFile(self, buildId, suffix):
        (packageResultPath, packageResultFile) = self._getPath(buildId, suffix)
        if os.path.isfile(packageResultFile):
            raise ArtifactExistsError()

        # open temporary file in destination directory
        if not os.path.isdir(packageResultPath):
            if self.__dirMode is not None:
                oldMask = os.umask(~self.__dirMode & 0o777)
            try:
                os.makedirs(packageResultPath, exist_ok=True)
            finally:
                if self.__dirMode is not None:
                    os.umask(oldMask)
        return LocalArchiveUploader(
            NamedTemporaryFile(dir=packageResultPath, delete=False),
            self.__fileMode,
            packageResultFile)

    def __uploadJenkins(self, step, buildIdFile, resultFile, suffix):
        """Generate upload shell script.

        We cannot simply copy the artifact to the final location as this is not
        atomic. Instead we create a temporary file at the repository root, copy
        the artifact there and hard-link the temporary file at the final
        location. If the link fails it is usually caused by a concurrent
        upload. Test that the artifact is readable in this case to distinguish
        it from other fatal errors.
        """
        if not self.canUploadJenkins():
            return ""

        if self.__dirMode is not None:
            setDirMode = "umask {:03o}".format(~self.__dirMode & 0o777)
        else:
            setDirMode = ":"

        if self.__fileMode is not None:
            setFileMode = 'chmod {:03o} "$T"'.format(self.__fileMode & 0o777)
        else:
            setFileMode = ""

        return "\n" + textwrap.dedent("""\
            # upload artifact
            cd $WORKSPACE
            BOB_UPLOAD_BID="$(hexdump -ve '/1 "%02x"' {BUILDID}){GEN}"
            BOB_UPLOAD_FILE={DIR}"/${{BOB_UPLOAD_BID:0:2}}/${{BOB_UPLOAD_BID:2:2}}/${{BOB_UPLOAD_BID:4}}{SUFFIX}"
            if [[ ! -e ${{BOB_UPLOAD_FILE}} ]] ; then
                (
                    set -eE
                    T="$(mktemp -p {DIR})"
                    trap 'rm -f $T' EXIT
                    cp {RESULT} "$T"
                    {SET_FILE_MODE}
                    ({SET_DIR_MODE} ; mkdir -p "${{BOB_UPLOAD_FILE%/*}}")
                    if ! ln -T "$T" "$BOB_UPLOAD_FILE" ; then
                        [[ -r "$BOB_UPLOAD_FILE" ]] || exit 2
                    fi
                ){FIXUP}
            fi""".format(DIR=quote(self.__basePath), BUILDID=quote(buildIdFile), RESULT=quote(resultFile),
                         FIXUP=" || echo Upload failed: $?" if self._ignoreErrors() else "",
                         GEN=ARCHIVE_GENERATION, SUFFIX=suffix,
                         SET_DIR_MODE=setDirMode, SET_FILE_MODE=setFileMode))

    def upload(self, step, buildIdFile, tgzFile):
        return self.__uploadJenkins(step, buildIdFile, tgzFile, ARTIFACT_SUFFIX)

    def download(self, step, buildIdFile, tgzFile):
        if not self.canDownloadJenkins():
            return ""

        return "\n" + textwrap.dedent("""\
            if [[ ! -e {RESULT} ]] ; then
                BOB_DOWNLOAD_BID="$(hexdump -ve '/1 "%02x"' {BUILDID}){GEN}"
                BOB_DOWNLOAD_FILE={DIR}"/${{BOB_DOWNLOAD_BID:0:2}}/${{BOB_DOWNLOAD_BID:2:2}}/${{BOB_DOWNLOAD_BID:4}}{SUFFIX}"
                cp "$BOB_DOWNLOAD_FILE" {RESULT} || echo Download failed: $?
            fi
            """.format(DIR=quote(self.__basePath), BUILDID=quote(buildIdFile), RESULT=quote(tgzFile),
                       GEN=ARCHIVE_GENERATION, SUFFIX=ARTIFACT_SUFFIX))

    def uploadJenkinsLiveBuildId(self, step, liveBuildId, buildId, isWin):
        return self.__uploadJenkins(step, liveBuildId, buildId, BUILDID_SUFFIX)

    def uploadJenkinsFingerprint(self, step, keyFile, fingerprintFile):
        return self.__uploadJenkins(step, keyFile, fingerprintFile, FINGERPRINT_SUFFIX)

class LocalArchiveDownloader:
    def __init__(self, name):
        try:
            self.fd = open(name, "rb")
        except OSError as e:
            raise ArtifactDownloadError(str(e))
    def __enter__(self):
        return (None, self.fd)
    def __exit__(self, exc_type, exc_value, traceback):
        self.fd.close()
        return False

class LocalArchiveUploader:
    def __init__(self, tmp, fileMode, destination):
        self.tmp = tmp
        self.fileMode = fileMode
        self.destination = destination
    def __enter__(self):
        return (None, self.tmp)
    def __exit__(self, exc_type, exc_value, traceback):
        self.tmp.close()
        # atomically move file to destination at end of upload
        if exc_type is None:
            if not isWindows():
                if self.fileMode is not None:
                    os.chmod(self.tmp.name, self.fileMode)
                # Cannot use os.rename() because it will unconditionally
                # replace an existing file. Instead we link the file at the
                # destination and unlink the temporary file.
                try:
                    os.link(self.tmp.name, self.destination)
                except FileExistsError:
                    pass # lost race
                finally:
                    os.unlink(self.tmp.name)
            else:
                try:
                    os.rename(self.tmp.name, self.destination)
                except OSError:
                    os.remove(self.tmp.name) # lost race
        else:
            os.unlink(self.tmp.name)
        return False


class SimpleHttpArchive(BaseArchive):
    def __init__(self, spec, secureSSL):
        super().__init__(spec)
        self.__url = urllib.parse.urlparse(spec["url"])
        self.__connection = None
        self.__sslVerify = spec.get("sslVerify", secureSSL)

    def __retry(self, request):
        retry = True
        while True:
            try:
                return (True, request())
            except (http.client.HTTPException, OSError) as e:
                self._resetConnection()
                if not retry: return (False, e)
                retry = False

    def _makeUrl(self, buildId, suffix):
        packageResultId = buildIdToName(buildId)
        return "/".join([self.__url.path, packageResultId[0:2], packageResultId[2:4],
            packageResultId[4:] + suffix])

    def _remoteName(self, buildId, suffix):
        url = self.__url
        return urllib.parse.urlunparse((url.scheme, url.netloc, self._makeUrl(buildId, suffix), '', '', ''))

    def _getConnection(self):
        if self.__connection is not None:
            return self.__connection

        url = self.__url
        if url.scheme == 'http':
            connection = http.client.HTTPConnection(url.hostname, url.port)
        elif url.scheme == 'https':
            ctx = None if self.__sslVerify else ssl.SSLContext(ssl.PROTOCOL_SSLv23)
            connection = http.client.HTTPSConnection(url.hostname, url.port,
                                                     context=ctx)
        else:
            raise BuildError("Unsupported URL scheme: '{}'".format(url.schema))

        self.__connection = connection
        return connection

    def _resetConnection(self):
        if self.__connection is not None:
            self.__connection.close()
            self.__connection = None

    def _openDownloadFile(self, buildId, suffix):
        (ok, result) = self.__retry(lambda: self.__openDownloadFile(buildId, suffix))
        if ok:
            return result
        else:
            raise ArtifactDownloadError(str(result))

    def __openDownloadFile(self, buildId, suffix):
        connection = self._getConnection()
        url = self._makeUrl(buildId, suffix)
        connection.request("GET", url,
                headers={ 'User-Agent' : 'BobBuildTool/{}'.format(BOB_VERSION) })
        response = connection.getresponse()
        if response.status == 200:
            return SimpleHttpDownloader(self, response)
        else:
            response.read()
            if response.status == 404:
                raise ArtifactNotFoundError()
            else:
                raise ArtifactDownloadError("{} {}".format(response.status,
                                                           response.reason))

    def _openUploadFile(self, buildId, suffix):
        (ok, result) = self.__retry(lambda: self.__openUploadFile(buildId, suffix))
        if ok:
            return result
        else:
            raise ArtifactUploadError(str(result))

    def __openUploadFile(self, buildId, suffix):
        connection = self._getConnection()
        url = self._makeUrl(buildId, suffix)

        # check if already there
        connection.request("HEAD", url,
                headers={ 'User-Agent' : 'BobBuildTool/{}'.format(BOB_VERSION) })
        response = connection.getresponse()
        response.read()
        if response.status == 200:
            raise ArtifactExistsError()
        elif response.status != 404:
            raise ArtifactUploadError("HEAD {} {}".format(response.status, response.reason))

        # create temporary file
        return SimpleHttpUploader(self, url)

    def _putUploadFile(self, url, tmp):
        (ok, result) = self.__retry(lambda: self.__putUploadFile(url, tmp))
        if ok:
            return result
        else:
            raise ArtifactUploadError(str(result))

    def __putUploadFile(self, url, tmp):
        # Determine file length outself and add a "Content-Length" header. This
        # used to work in Python 3.5 automatically but was removed later.
        tmp.seek(0, os.SEEK_END)
        length = str(tmp.tell())
        tmp.seek(0)
        connection = self._getConnection()
        connection.request("PUT", url, tmp, headers={ 'Content-Length' : length,
            'If-None-Match' : '*', 'User-Agent' : 'BobBuildTool/{}'.format(BOB_VERSION) })
        response = connection.getresponse()
        response.read()
        if response.status == 412:
            # precondition failed -> lost race with other upload
            raise ArtifactExistsError()
        elif response.status not in [200, 201, 204]:
            raise ArtifactUploadError("PUT {} {}".format(response.status, response.reason))

    def __uploadJenkins(self, step, keyFile, contentFile, suffix):
        # only upload if requested
        if not self.canUploadJenkins():
            return ""

        # upload with curl if file does not exist yet on server
        insecure = "" if self.__sslVerify else "-k"
        return "\n" + textwrap.dedent("""\
            # upload artifact
            cd $WORKSPACE
            BOB_UPLOAD_BID="$(hexdump -ve '/1 "%02x"' {KEYFILE}){GEN}"
            BOB_UPLOAD_URL={URL}"/${{BOB_UPLOAD_BID:0:2}}/${{BOB_UPLOAD_BID:2:2}}/${{BOB_UPLOAD_BID:4}}{SUFFIX}"
            if ! curl --output /dev/null --silent --head --fail {INSECURE} "$BOB_UPLOAD_URL" ; then
                BOB_UPLOAD_RSP=$(curl -sSgf {INSECURE} -w '%{{http_code}}' -H 'If-None-Match: *' -T {CONTENTFILE} "$BOB_UPLOAD_URL" || true)
                if [[ $BOB_UPLOAD_RSP != 2?? && $BOB_UPLOAD_RSP != 412 ]]; then
                    echo "Upload failed with code $BOB_UPLOAD_RSP"{FAIL}
                fi
            fi""".format(URL=quote(self.__url.geturl()), KEYFILE=quote(keyFile),
                         CONTENTFILE=quote(contentFile),
                         FAIL="" if self._ignoreErrors() else "; exit 1",
                         GEN=ARCHIVE_GENERATION, SUFFIX=suffix,
                         INSECURE=insecure))

    def upload(self, step, buildIdFile, tgzFile):
        return self.__uploadJenkins(step, buildIdFile, tgzFile, ARTIFACT_SUFFIX)

    def download(self, step, buildIdFile, tgzFile):
        # only download if requested
        if not self.canDownloadJenkins():
            return ""

        insecure = "" if self.__sslVerify else "-k"
        return "\n" + textwrap.dedent("""\
            if [[ ! -e {RESULT} ]] ; then
                BOB_DOWNLOAD_BID="$(hexdump -ve '/1 "%02x"' {BUILDID}){GEN}"
                BOB_DOWNLOAD_URL={URL}"/${{BOB_DOWNLOAD_BID:0:2}}/${{BOB_DOWNLOAD_BID:2:2}}/${{BOB_DOWNLOAD_BID:4}}{SUFFIX}"
                curl -sSg {INSECURE} --fail -o {RESULT} "$BOB_DOWNLOAD_URL" || echo Download failed: $?
            fi
            """.format(URL=quote(self.__url.geturl()), BUILDID=quote(buildIdFile), RESULT=quote(tgzFile),
                       GEN=ARCHIVE_GENERATION, SUFFIX=ARTIFACT_SUFFIX,
                       INSECURE=insecure))

    def uploadJenkinsLiveBuildId(self, step, liveBuildId, buildId, isWin):
        return self.__uploadJenkins(step, liveBuildId, buildId, BUILDID_SUFFIX)

    def uploadJenkinsFingerprint(self, step, keyFile, fingerprintFile):
        return self.__uploadJenkins(step, keyFile, fingerprintFile, FINGERPRINT_SUFFIX)

class SimpleHttpDownloader:
    def __init__(self, archiver, response):
        self.archiver = archiver
        self.response = response
    def __enter__(self):
        return (None, self.response)
    def __exit__(self, exc_type, exc_value, traceback):
        # reset connection on abnormal termination
        if exc_type is not None:
            self.archiver._resetConnection()
        return False

class SimpleHttpUploader:
    def __init__(self, archiver, url):
        self.archiver = archiver
        self.tmp = TemporaryFile()
        self.url = url
    def __enter__(self):
        return (None, self.tmp)
    def __exit__(self, exc_type, exc_value, traceback):
        try:
            # do actual upload on regular handle close
            if exc_type is None:
                self.archiver._putUploadFile(self.url, self.tmp)
        finally:
            self.tmp.close()
        return False


class CustomArchive(BaseArchive):
    """Custom command archive"""

    def __init__(self, spec, whiteList):
        super().__init__(spec)
        self.__downloadCmd = spec.get("download")
        self.__uploadCmd = spec.get("upload")
        self.__whiteList = whiteList

    def _makeUrl(self, buildId, suffix):
        packageResultId = buildIdToName(buildId)
        return "/".join([packageResultId[0:2], packageResultId[2:4],
            packageResultId[4:] + suffix])

    def _remoteName(self, buildId, suffix):
        return self._makeUrl(buildId, suffix)

    def canDownloadLocal(self):
        return super().canDownloadLocal() and (self.__downloadCmd is not None)

    def canUploadLocal(self):
        return super().canUploadLocal() and (self.__uploadCmd is not None)

    def canDownloadJenkins(self):
        return super().canDownloadJenkins() and (self.__downloadCmd is not None)

    def canUploadJenkins(self):
        return super().canUploadJenkins() and (self.__uploadCmd is not None)

    def _openDownloadFile(self, buildId, suffix):
        (tmpFd, tmpName) = mkstemp()
        url = self._makeUrl(buildId, suffix)
        try:
            os.close(tmpFd)
            env = { k:v for (k,v) in os.environ.items() if k in self.__whiteList }
            env["BOB_LOCAL_ARTIFACT"] = tmpName
            env["BOB_REMOTE_ARTIFACT"] = url
            ret = subprocess.call(["/bin/bash", "-ec", self.__downloadCmd],
                stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL,
                cwd="/tmp", env=env)
            if ret == 0:
                ret = tmpName
                tmpName = None
                return CustomDownloader(ret)
            else:
                raise ArtifactDownloadError("failed (exit {})".format(ret))
        finally:
            if tmpName is not None: os.unlink(tmpName)

    def _openUploadFile(self, buildId, suffix):
        (tmpFd, tmpName) = mkstemp()
        os.close(tmpFd)
        return CustomUploader(tmpName, self._makeUrl(buildId, suffix), self.__whiteList,
            self.__uploadCmd)

    def __uploadJenkins(self, step, buildIdFile, tgzFile, suffix):
        # only upload if requested
        if not self.canUploadJenkins():
            return ""

        cmd = self.__uploadCmd
        if self._ignoreErrors():
            # wrap in subshell
            cmd = "( " + cmd + " ) || echo Upload failed: $?"

        return "\n" + textwrap.dedent("""\
            # upload artifact
            cd $WORKSPACE
            BOB_UPLOAD_BID="$(hexdump -ve '/1 "%02x"' {BUILDID}){GEN}"
            BOB_LOCAL_ARTIFACT={RESULT}
            BOB_REMOTE_ARTIFACT="${{BOB_UPLOAD_BID:0:2}}/${{BOB_UPLOAD_BID:2:2}}/${{BOB_UPLOAD_BID:4}}{SUFFIX}"
            """.format(BUILDID=quote(buildIdFile), RESULT=quote(tgzFile),
                       GEN=ARCHIVE_GENERATION, SUFFIX=suffix)) + cmd

    def upload(self, step, buildIdFile, tgzFile):
        return self.__uploadJenkins(step, buildIdFile, tgzFile, ARTIFACT_SUFFIX)

    def download(self, step, buildIdFile, tgzFile):
        # only download if requested
        if not self.canDownloadJenkins():
            return ""

        return """
if [[ ! -e {RESULT} ]] ; then
    BOB_DOWNLOAD_BID="$(hexdump -ve '/1 "%02x"' {BUILDID}){GEN}"
    BOB_LOCAL_ARTIFACT={RESULT}
    BOB_REMOTE_ARTIFACT="${{BOB_DOWNLOAD_BID:0:2}}/${{BOB_DOWNLOAD_BID:2:2}}/${{BOB_DOWNLOAD_BID:4}}{SUFFIX}"
    {CMD}
fi
""".format(CMD=self.__downloadCmd, BUILDID=quote(buildIdFile), RESULT=quote(tgzFile),
           GEN=ARCHIVE_GENERATION, SUFFIX=ARTIFACT_SUFFIX)

    def uploadJenkinsLiveBuildId(self, step, liveBuildId, buildId, isWin):
        return self.__uploadJenkins(step, liveBuildId, buildId, BUILDID_SUFFIX)

    def uploadJenkinsFingerprint(self, step, keyFile, fingerprintFile):
        return self.__uploadJenkins(step, keyFile, fingerprintFile, FINGERPRINT_SUFFIX)

class CustomDownloader:
    def __init__(self, name):
        self.name = name
    def __enter__(self):
        return (self.name, None)
    def __exit__(self, exc_type, exc_value, traceback):
        os.unlink(self.name)
        return False

class CustomUploader:
    def __init__(self, name, remoteName, whiteList, uploadCmd):
        self.name = name
        self.remoteName = remoteName
        self.whiteList = whiteList
        self.uploadCmd = uploadCmd

    def __enter__(self):
        return (self.name, None)

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            if exc_type is None:
                env = { k:v for (k,v) in os.environ.items() if k in self.whiteList }
                env["BOB_LOCAL_ARTIFACT"] = self.name
                env["BOB_REMOTE_ARTIFACT"] = self.remoteName
                ret = subprocess.call(["/bin/bash", "-ec", self.uploadCmd],
                    stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL,
                    cwd="/tmp", env=env)
                if ret != 0:
                    raise ArtifactUploadError("command return with status {}".format(ret))
        finally:
            os.unlink(self.name)
        return False


class AzureArchive(BaseArchive):
    def __init__(self, spec):
        super().__init__(spec)
        self.__container = spec['container']
        self.__account = spec['account']
        self.__key = spec.get('key')
        self.__sasToken = spec.get('sasToken')
        try:
            from azure.storage.blob import BlockBlobService
        except ImportError:
            raise BuildError("azure-storage-blob Python3 library not installed!")
        self.__service = BlockBlobService(account_name=self.__account,
            account_key=self.__key, sas_token=self.__sasToken, socket_timeout=6000)

    @staticmethod
    def __makeBlobName(buildId, suffix):
        packageResultId = buildIdToName(buildId)
        return "/".join([packageResultId[0:2], packageResultId[2:4],
            packageResultId[4:] + suffix])

    def _remoteName(self, buildId, suffix):
        return "https://{}.blob.core.windows.net/{}/{}".format(self.__account,
            self.__container, self.__makeBlobName(buildId, suffix))

    def _openDownloadFile(self, buildId, suffix):
        from azure.common import AzureException, AzureMissingResourceHttpError
        (tmpFd, tmpName) = mkstemp()
        try:
            os.close(tmpFd)
            self.__service.get_blob_to_path(self.__container,
                self.__makeBlobName(buildId, suffix), tmpName)
            ret = tmpName
            tmpName = None
            return AzureDownloader(ret)
        except AzureMissingResourceHttpError:
            raise ArtifactNotFoundError()
        except AzureException as e:
            raise ArtifactDownloadError(str(e))
        finally:
            if tmpName is not None: os.unlink(tmpName)

    def _openUploadFile(self, buildId, suffix):
        from azure.common import AzureException

        blobName = self.__makeBlobName(buildId, suffix)
        try:
            if self.__service.exists(self.__container, blobName):
                raise ArtifactExistsError()
        except AzureException as e:
            raise ArtifactUploadError(str(e))
        (tmpFd, tmpName) = mkstemp()
        os.close(tmpFd)
        return AzureUploader(self.__service, self.__container, tmpName, blobName)

    def __uploadJenkins(self, step, keyFile, contentFile, suffix):
        if not self.canUploadJenkins():
            return ""

        args = []
        if self.__key: args.append("--key=" + self.__key)
        if self.__sasToken: args.append("--sas-token=" + self.__sasToken)

        return "\n" + textwrap.dedent("""\
            # upload artifact
            cd $WORKSPACE
            bob _upload azure {ARGS} {ACCOUNT} {CONTAINER} {KEYFILE} {SUFFIX} {CONTENTFILE}{FIXUP}
            """.format(ARGS=" ".join(map(quote, args)), ACCOUNT=quote(self.__account),
                       CONTAINER=quote(self.__container), KEYFILE=quote(keyFile),
                       CONTENTFILE=quote(contentFile),
                       FIXUP=" || echo Upload failed: $?" if self._ignoreErrors() else "",
                       SUFFIX=suffix))

    def upload(self, step, buildIdFile, tgzFile):
        return self.__uploadJenkins(step, buildIdFile, tgzFile, ARTIFACT_SUFFIX)

    def download(self, step, buildIdFile, tgzFile):
        if not self.canDownloadJenkins():
            return ""

        args = []
        if self.__key: args.append("--key=" + self.__key)
        if self.__sasToken: args.append("--sas-token=" + self.__sasToken)

        return "\n" + textwrap.dedent("""\
            if [[ ! -e {RESULT} ]] ; then
                bob _download azure {ARGS} {ACCOUNT} {CONTAINER} {BUILDID} {SUFFIX} {RESULT} || echo Download failed: $?
            fi
            """.format(ARGS=" ".join(map(quote, args)), ACCOUNT=quote(self.__account),
                       CONTAINER=self.__container, BUILDID=quote(buildIdFile),
                       RESULT=quote(tgzFile), SUFFIX=ARTIFACT_SUFFIX))

    def uploadJenkinsLiveBuildId(self, step, liveBuildId, buildId, isWin):
        return self.__uploadJenkins(step, liveBuildId, buildId, BUILDID_SUFFIX)

    def uploadJenkinsFingerprint(self, step, keyFile, fingerprintFile):
        return self.__uploadJenkins(step, keyFile, fingerprintFile, FINGERPRINT_SUFFIX)

    @staticmethod
    def scriptDownload(args):
        service, container, remoteBlob, localFile = AzureArchive.scriptGetService(args)
        from azure.common import AzureException

        # Download into temporary file and rename if downloaded successfully
        tmpName = None
        try:
            (tmpFd, tmpName) = mkstemp(dir=".")
            os.close(tmpFd)
            service.get_blob_to_path(container, remoteBlob, tmpName)
            os.rename(tmpName, localFile)
            tmpName = None
        except (OSError, AzureException) as e:
            raise BuildError("Download failed: " + str(e))
        finally:
            if tmpName is not None: os.unlink(tmpName)

    @staticmethod
    def scriptUpload(args):
        service, container, remoteBlob, localFile = AzureArchive.scriptGetService(args)
        from azure.common import AzureException, AzureConflictHttpError
        try:
            service.create_blob_from_path(container, remoteBlob, localFile, if_none_match="*")
            print("OK")
        except AzureConflictHttpError:
            print("skipped")
        except (OSError, AzureException) as e:
            raise BuildError("Upload failed: " + str(e))

    @staticmethod
    def scriptGetService(args):
        parser = argparse.ArgumentParser()
        parser.add_argument('account')
        parser.add_argument('container')
        parser.add_argument('buildid')
        parser.add_argument('suffix')
        parser.add_argument('file')
        parser.add_argument('--key')
        parser.add_argument('--sas-token')
        args = parser.parse_args(args)

        try:
            from azure.storage.blob import BlockBlobService
        except ImportError:
            raise BuildError("azure-storage-blob Python3 library not installed!")

        service = BlockBlobService(account_name=args.account, account_key=args.key,
            sas_token=args.sas_token, socket_timeout=6000)

        try:
            with open(args.buildid, 'rb') as f:
                remoteBlob = AzureArchive.__makeBlobName(f.read(), args.suffix)
        except OSError as e:
            raise BuildError(str(e))

        return (service, args.container, remoteBlob, args.file)

class AzureDownloader:
    def __init__(self, name):
        self.name = name
    def __enter__(self):
        return (self.name, None)
    def __exit__(self, exc_type, exc_value, traceback):
        os.unlink(self.name)
        return False

class AzureUploader:
    def __init__(self, service, container, name, remoteName):
        self.__service = service
        self.__container = container
        self.__name = name
        self.__remoteName = remoteName

    def __enter__(self):
        return (self.__name, None)

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            if exc_type is None:
                self.__upload()
        finally:
            os.unlink(self.__name)
        return False

    def __upload(self):
        from azure.common import AzureException, AzureConflictHttpError
        try:
            self.__service.create_blob_from_path(self.__container,
                self.__remoteName, self.__name, if_none_match="*")
        except AzureConflictHttpError:
            raise ArtifactExistsError()
        except AzureException as e:
            raise ArtifactUploadError(str(e))


class MultiArchive:
    def __init__(self, archives):
        self.__archives = archives

    def wantDownload(self, enable):
        for i in self.__archives: i.wantDownload(enable)

    def wantUpload(self, enable):
        for i in self.__archives: i.wantUpload(enable)

    def canDownloadLocal(self):
        return any(i.canDownloadLocal() for i in self.__archives)

    def canUploadLocal(self):
        return any(i.canUploadLocal() for i in self.__archives)

    def canDownloadJenkins(self):
        return any(i.canDownloadJenkins() for i in self.__archives)

    def canUploadJenkins(self):
        return any(i.canUploadJenkins() for i in self.__archives)

    async def uploadPackage(self, step, buildId, audit, content):
        for i in self.__archives:
            if not i.canUploadLocal(): continue
            await i.uploadPackage(step, buildId, audit, content)

    async def downloadPackage(self, step, buildId, audit, content):
        for i in self.__archives:
            if not i.canDownloadLocal(): continue
            if await i.downloadPackage(step, buildId, audit, content): return True
        return False

    def upload(self, step, buildIdFile, tgzFile):
        return "\n".join(
            i.upload(step, buildIdFile, tgzFile) for i in self.__archives
            if i.canUploadJenkins())

    def download(self, step, buildIdFile, tgzFile):
        return "\n".join(
            i.download(step, buildIdFile, tgzFile) for i in self.__archives
            if i.canDownloadJenkins())

    async def uploadLocalLiveBuildId(self, step, liveBuildId, buildId):
        for i in self.__archives:
            if not i.canUploadLocal(): continue
            await i.uploadLocalLiveBuildId(step, liveBuildId, buildId)

    async def downloadLocalLiveBuildId(self, step, liveBuildId):
        ret = None
        for i in self.__archives:
            if not i.canDownloadLocal(): continue
            ret = await i.downloadLocalLiveBuildId(step, liveBuildId)
            if ret is not None: break
        return ret

    def uploadJenkinsLiveBuildId(self, step, liveBuildId, buildId, isWin):
        return "\n".join(
            i.uploadJenkinsLiveBuildId(step, liveBuildId, buildId, isWin)
            for i in self.__archives if i.canUploadJenkins())

    async def uploadLocalFingerprint(self, step, key, fingerprint):
        for i in self.__archives:
            if not i.canUploadLocal(): continue
            await i.uploadLocalFingerprint(step, key, fingerprint)

    async def downloadLocalFingerprint(self, step, key):
        ret = None
        for i in self.__archives:
            if not i.canDownloadLocal(): continue
            ret = await i.downloadLocalFingerprint(step, key)
            if ret is not None: break
        return ret

    def uploadJenkinsFingerprint(self, step, keyFile, fingerprintFile):
        return "\n".join(
            i.uploadJenkinsFingerprint(step, keyFile, fingerprintFile)
            for i in self.__archives if i.canUploadJenkins())


def getSingleArchiver(recipes, archiveSpec):
    archiveBackend = archiveSpec.get("backend", "none")
    if archiveBackend == "file":
        return LocalArchive(archiveSpec)
    elif archiveBackend == "http":
        return SimpleHttpArchive(archiveSpec, recipes.getPolicy('secureSSL'))
    elif archiveBackend == "shell":
        return CustomArchive(archiveSpec, recipes.envWhiteList())
    elif archiveBackend == "azure":
        return AzureArchive(archiveSpec)
    elif archiveBackend == "none":
        return DummyArchive()
    else:
        raise BuildError("Invalid archive backend: "+archiveBackend)

def getArchiver(recipes):
    archiveSpec = recipes.archiveSpec()
    if isinstance(archiveSpec, list):
        return MultiArchive([ getSingleArchiver(recipes, i) for i in archiveSpec ])
    else:
        return getSingleArchiver(recipes, archiveSpec)

def doDownload(args, bobRoot):
    archiveBackend = args[0]
    if archiveBackend == "azure":
        AzureArchive.scriptDownload(args[1:])
    else:
        raise BuildError("Invalid archive backend: "+archiveBackend)

def doUpload(args, bobRoot):
    archiveBackend = args[0]
    if archiveBackend == "azure":
        AzureArchive.scriptUpload(args[1:])
    else:
        raise BuildError("Invalid archive backend: "+archiveBackend)



# Bob build tool
# Copyright (C) 2016  TechniSat Digital GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later

from .errors import BuildError, ParseError
from binascii import hexlify
from tempfile import NamedTemporaryFile, TemporaryDirectory
import collections.abc
import hashlib
import logging
import os
import re
import shutil
import stat
import struct
import sys

def hashString(string):
    h = hashlib.md5()
    h.update(string.encode("utf8"))
    return h.digest()

def asHexStr(binary):
    return hexlify(binary).decode("ascii")

def joinLines(*lines):
    return "\n".join(l for l in lines if l)

def joinScripts(scripts, glue):
    scripts = [ s for s in scripts if ((s is not None) and (s != "")) ]
    if scripts != []:
        return glue.join(scripts)
    else:
        return None

def sliceString(data, chunk):
    """Return iterator that slices string in "chunk" size strings."""
    def genSlice(i = 0):
        r = data[i:i+chunk]
        while len(r) > 0:
            yield r
            i += chunk
            r = data[i:i+chunk]
    return iter(genSlice())

def quotePwsh(string):
    """Create a PowerShell string literal"""
    return "'" + string.replace("'", "''") + "'"

def escapePwsh(string):
    """Escape a string so that no meta characters are interpreted by PowerShell"""
    return string.replace('"', '`"').replace('$', '`$')

def quoteCmdExe(string):
    """Quote a string for cmd.exe to prevent interpretation of meta characters"""
    if any(c in string for c in " \"()[]{}^=;!'+,`~"):
        return '"' + string.replace('"', '""') + '"'
    else:
        return string

def removePath(path):
    if sys.platform == "win32":
        def onerror(func, path, exc):
            os.chmod(path, stat.S_IWRITE)
            os.unlink(path)
    else:
        onerror = None

    try:
        if os.path.lexists(path):
            if os.path.isdir(path) and not os.path.islink(path):
                shutil.rmtree(path, onerror=onerror)
            else:
                os.unlink(path)
    except OSError as e:
        raise BuildError("Error removing '"+path+"': " + str(e))

def emptyDirectory(path):
    try:
        if os.path.exists(path):
            for f in os.listdir(path): removePath(os.path.join(path, f))
    except OSError as e:
        raise BuildError("Error cleaning '"+path+"': " + str(e))

# Recursively merge entries of two dictonaries.
#
# Expect that both arguments have the same schema. Dictionaries are merged
# key-by-key. Lists are appended. Returns merged result.
#
# See: http://stackoverflow.com/questions/3232943
def updateDicRecursive(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = updateDicRecursive(d.get(k, {}), v)
        elif isinstance(v, list):
            d[k] = d.get(k, []) + v
        else:
            d[k] = u[k]
    return d

# Compare PEP 440 versions. Not strictly according to spec but enough for us.
def compareVersion(origLeft, origRight):

    def cmp(l, r):
        if (len(l) == 0) and (len(r) == 0): return 0
        if len(l) == 0: l = ["0"]
        if len(r) == 0: r = ["0"]
        if int(l[0]) < int(r[0]):
            return -1
        elif int(l[0]) > int(r[0]):
            return 1
        else:
            return cmp(l[1:], r[1:])

    try:
        r = re.compile(r"^(?P<version>[0-9]+(?:\.[0-9]+){0,2})(?:rc(?P<rc>[0-9]+))?(?:.dev(?P<dist>[0-9]+))?(?:\+.*)?$")
        left = r.match(origLeft).groupdict()
        right = r.match(origRight).groupdict()

        # Compare version number. If an element is missing it is assumed to be 0.
        ret = cmp(left["version"].split("."), right["version"].split("."))

        # If both versions are equal than the higher release candidate wins. A
        # version without release candidate is considered more higher.
        if ret == 0:
            lrc = 9999 if  left["rc"] is None else int( left["rc"])
            rrc = 9999 if right["rc"] is None else int(right["rc"])
            if lrc < rrc:
                ret = -1
            elif lrc > rrc:
                ret = 1

        # If we still have a tie then the smallest distance wins.
        if ret == 0:
            ldist = 0xFFFF if  left["dist"] is None else int( left["dist"])
            rdist = 0xFFFF if right["dist"] is None else int(right["dist"])
            if ldist < rdist:
                ret = -1
            elif ldist > rdist:
                ret = 1

    except Exception:
        raise ParseError("Cannot compare version numbers ('{}' vs. '{}'): bad format!"
                            .format(origLeft, origRight))
    return ret

def isWindows():
    """Check if we run on a windows platform.

    We have to rule out MSYS(2) and Cygwin as they are advertised a POSIX but
    in fact cannot truly hide the underlying system.
    """
    if os.name == 'posix':
        p = sys.platform
        if p.startswith('msys'): return True
        if p.startswith('cygwin'): return True
        return False
    return True

if isWindows():
    INVALID_CHAR_TRANS = str.maketrans(':*?<>"|', '_______')
else:
    INVALID_CHAR_TRANS = str.maketrans('', '')


__platformTag = None

def getPlatformTag():
    # cached on first call
    global __platformTag
    if __platformTag is not None:
        return __platformTag

    p = sys.platform
    if p == 'win32':
        ret = b'w'
    elif p in ('msys', 'cygwin'):
        ret = b'm'
    else:
        ret = b''

    # On Windows it depends on the SeCreateSymbolicLinkPrivilege capability if
    # it is possible to create symlinks. Try to create a symlink to see if we
    # have the privilege. Either the symlink() call fails directly or MSYS
    # silently creates a copy (unless MSYS=winsymlinks:nativestrict is set).
    if p in ('msys', 'cygwin', 'win32'):
        canSymlink = False
        try:
            with TemporaryDirectory() as tmp:
                with open(os.path.join(tmp, "file"), "w") as f:
                    pass
                canary = os.path.join(tmp, "canary")
                os.symlink("file", canary)
                canSymlink = os.path.islink(canary)
        except OSError:
            pass
        if canSymlink:
            ret += b'l'

    # cache result
    __platformTag = ret
    return ret

### directory hashing ###

def hashFile(path, hasher=hashlib.sha1):
    m = hasher()
    try:
        with open(path, 'rb', buffering=0) as f:
            buf = f.read(16384)
            while len(buf) > 0:
                m.update(buf)
                buf = f.read(16384)
    except OSError as e:
        logging.getLogger(__name__).warning("Cannot hash file: %s", str(e))
    return m.digest()

def float2ns(v):
    return int(v * 1000000000)

class DirHasher:
    IGNORE_DIRS = frozenset([
        os.fsencode(".git"),
        os.fsencode(".portage-cache"),
        os.fsencode(".svn"),
    ])
    IGNORE_FILES = frozenset([
        os.fsencode("BaseDirList.txt"),
    ])

    class FileIndex:
        SIGNATURE        = b'BOB1'
        CACHE_ENTRY_FMT  = '=QQLQLQ20sH'
        CACHE_ENTRY_SIZE = struct.calcsize(CACHE_ENTRY_FMT)

        class Stat:
            def __init__(self):
                self.name = b""
                self.ctime = 0
                self.mtime = 0
                self.dev = 0
                self.ino = 0
                self.mode = 0
                self.size = 0
                self.digest = b''

            def __repr__(self):
                return "Stat(name={}, ctime={}, mtime={}, dev={}, ino={}, mode={}, size={}, digest={})".format(
                    self.name, self.ctime, self.mtime, self.dev, self.ino, self.mode, self.size, self.digest)

        def __init__(self, cachePath):
            self.__cachePath = cachePath
            self.__cacheDir = os.path.dirname(cachePath)

        def open(self):
            self.__inPos = 0
            self.__inPosOld = 0
            self.__outFile = None
            self.__current = DirHasher.FileIndex.Stat()
            try:
                if os.path.exists(self.__cachePath):
                    self.__inFile = open(self.__cachePath, "rb")
                    sig = self.__inFile.read(4)
                    if sig == DirHasher.FileIndex.SIGNATURE:
                        self.__mismatch = False
                        self.__inPos = self.__inPosOld = 4
                        self.__readEntry() # prefetch first entry
                    else:
                        logging.getLogger(__name__).info(
                            "Wrong signature at '%s': %s", self.__cachePath, sig)
                        self.__inFile.close()
                        self.__inFile = None
                        self.__mismatch = True
                else:
                    self.__inFile = None
                    self.__mismatch = True
            except OSError as e:
                raise BuildError("Error opening hash cache: " + str(e))

        def close(self):
            try:
                if self.__inFile:
                    self.__inFile.close()
                if self.__outFile:
                    self.__outFile.close()
                    os.replace(self.__outFile.name, self.__cachePath)
            except OSError as e:
                raise BuildError("Error closing hash cache: " + str(e))

        def __readEntry(self):
            if not self.__inFile: return False
            raw = self.__inFile.read(DirHasher.FileIndex.CACHE_ENTRY_SIZE)
            if len(raw) < DirHasher.FileIndex.CACHE_ENTRY_SIZE: return False
            e = self.__current
            (e.ctime, e.mtime, e.dev, e.ino, e.mode, e.size, e.digest,
                nameLen) = struct.unpack(DirHasher.FileIndex.CACHE_ENTRY_FMT, raw)
            e.name = self.__inFile.read(nameLen)
            self.__inPosOld = self.__inPos
            self.__inPos += DirHasher.FileIndex.CACHE_ENTRY_SIZE + nameLen
            return True

        def __writeEntry(self, name, st, digest):
            if not self.__outFile:
                self.__outFile = NamedTemporaryFile(mode="wb", dir=self.__cacheDir, delete=False)
                if self.__inFile:
                    pos = self.__inFile.tell()
                    self.__inFile.seek(0)
                    self.__outFile.write(self.__inFile.read(self.__inPosOld))
                    self.__inFile.seek(pos)
                else:
                    self.__outFile.write(DirHasher.FileIndex.SIGNATURE)
            self.__outFile.write(struct.pack(DirHasher.FileIndex.CACHE_ENTRY_FMT, float2ns(st.st_ctime),
                float2ns(st.st_mtime), st.st_dev, st.st_ino, st.st_mode, st.st_size,
                digest, len(name)))
            self.__outFile.write(name)

        def __match(self, name, st):
            while self.__current.name < name:
                if not self.__readEntry(): break
            e = self.__current
            res = ((e.name == name) and (e.ctime == float2ns(st.st_ctime)) and
                (e.mtime == float2ns(st.st_mtime)) and (e.dev == st.st_dev) and
                (e.ino == st.st_ino) and (e.mode == st.st_mode) and
                (e.size == st.st_size))
            #if not res: print("Mismatch", e.name, name, e, st)
            return res

        def check(self, prefix, name, st, process):
            if self.__match(name, st):
                digest = self.__current.digest
            else:
                digest = process(os.path.join(prefix, name) if name else prefix)
                self.__mismatch = True
            if self.__mismatch:
                self.__writeEntry(name, st, digest)
            return digest

    class NullIndex:
        def __init__(self):
            pass

        def open(self):
            pass

        def close(self):
            pass

        def check(self, prefix, name, st, process):
            return process(os.path.join(prefix, name) if name else prefix)

    def __init__(self, basePath=None, ignoreDirs=None):
        if basePath:
            self.__index = DirHasher.FileIndex(basePath)
        else:
            self.__index = DirHasher.NullIndex()
        if ignoreDirs:
            self.__ignoreDirs = DirHasher.IGNORE_DIRS | frozenset(os.fsencode(i) for i in ignoreDirs)
        else:
            self.__ignoreDirs = DirHasher.IGNORE_DIRS

    def __hashEntry(self, prefix, entry, s):
        if stat.S_ISREG(s.st_mode):
            digest = self.__index.check(prefix, entry, s, hashFile)
        elif stat.S_ISDIR(s.st_mode):
            digest = self.__hashDir(prefix, entry)
        elif stat.S_ISLNK(s.st_mode):
            digest = self.__index.check(prefix, entry, s, DirHasher.__hashLink)
        elif stat.S_ISBLK(s.st_mode) or stat.S_ISCHR(s.st_mode):
            digest = struct.pack("<L", s.st_rdev)
        elif stat.S_ISFIFO(s.st_mode):
            digest = b''
        else:
            digest = b''
            logging.getLogger(__name__).warning("Unknown file: %s", entry)

        return digest

    if (sys.platform == "win32") and (sys.version_info.minor <= 7):
        # Before Python 3.8 the os.readlink method did not accept byte encoded
        # file names on Windows.
        @staticmethod
        def __hashLink(path):
            m = hashlib.sha1()
            try:
                m.update(os.fsencode(os.readlink(os.fsdecode(path))))
            except OSError as e:
                logging.getLogger(__name__).warning("Cannot hash link: %s", str(e))
            return m.digest()
    else:
        @staticmethod
        def __hashLink(path):
            m = hashlib.sha1()
            try:
                m.update(os.readlink(path))
            except OSError as e:
                logging.getLogger(__name__).warning("Cannot hash link: %s", str(e))
            return m.digest()

    def __hashDir(self, prefix, path=b''):
        entries = []
        try:
            dirEntries = os.listdir(os.path.join(prefix, path if path else b'.'))
        except OSError as e:
            logging.getLogger(__name__).warning("Cannot list directory: %s", str(e))
            dirEntries = []

        for f in dirEntries:
            e = os.path.join(path, f)
            try:
                s = os.lstat(os.path.join(prefix, e))
                if stat.S_ISDIR(s.st_mode):
                    # skip useless directories
                    if f in self.__ignoreDirs: continue
                    # add training '/' for directores for correct sorting
                    f = f + os.fsencode(os.path.sep)
                else:
                    # skip useless files
                    if f in DirHasher.IGNORE_FILES: continue
                entries.append((e, f, s))
            except OSError as err:
                logging.getLogger(__name__).warning("Cannot stat '%s': %s", e, str(err))
        entries = sorted(entries, key=lambda x: x[1])
        dirList = [
            (struct.pack("=L", s.st_mode) + self.__hashEntry(prefix, e, s) + f)
            for (e, f, s) in entries
        ]
        dirBlob = b"".join(dirList)
        m = hashlib.sha1()
        m.update(dirBlob)
        return m.digest()

    def hashDirectory(self, path):
        self.__index.open()
        try:
            return self.__hashDir(os.fsencode(path))
        finally:
            self.__index.close()

    def hashPath(self, path):
        path = os.fsencode(path)
        try:
            s = os.lstat(path)
        except OSError as err:
            logging.getLogger(__name__).warning("Cannot stat '%s': %s", path, str(err))
            return b''

        self.__index.open()
        try:
            return self.__hashEntry(path, b'', s)
        finally:
            self.__index.close()


def hashDirectory(path, index=None, ignoreDirs=None):
    return DirHasher(index, ignoreDirs).hashDirectory(path)

def hashPath(path, index=None, ignoreDirs=None):
    return DirHasher(index, ignoreDirs).hashPath(path)

def binStat(path):
    st = os.stat(path)
    return struct.pack('=QQLqLQ', float2ns(st.st_ctime), float2ns(st.st_mtime),
                       st.st_dev, st.st_ino, st.st_mode, st.st_size)


# There are two "magic" modules with similar functionality. Find out which one we got and adapt.
def summonMagic():
    import magic
    if hasattr(magic, 'from_file'):
        # https://pypi.python.org/pypi/python-magic
        return magic
    elif hasattr(magic, 'open'):
        # http://www.darwinsys.com/file/, in Debian as python3-magic
        class WrapMagic:
            def __init__(self):
                self.magic = magic.open(magic.NONE)
                self.magic.load()

            def from_file(self, name):
                return self.magic.file(name)
        return WrapMagic()
    else:
        raise NotImplementedError("I do not understand your magic")

### directory copy ###

def copyTree(src, dst):
    try:
        names = os.listdir(src)
        os.makedirs(dst, exist_ok=True)
    except OSError as e:
        logging.getLogger(__name__).error("Copy '%s' to '%s' failed: %s", src,
                                          dst, str(e))
        return False

    ret = True
    for name in names:
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if os.path.isdir(srcname):
                if os.path.exists(dstname) and not os.path.isdir(dstname):
                    logging.getLogger(__name__).error(
                        "Cannon overwrite non-directory '%s' with directory '%s'",
                        dstname, srcname)
                    ret = False
                else:
                    ret = copyTree(srcname, dstname) and ret
            else:
                if os.path.lexists(dstname):
                    os.unlink(dstname)
                if os.path.islink(srcname):
                    linkto = os.readlink(srcname)
                    os.symlink(linkto, dstname)
                else:
                    shutil.copy(srcname, dstname)
        except OSError as e:
            logging.getLogger(__name__).error("Copy failed: %s", str(e))
            ret = False

    return ret


def processDefines(defs):
    defines = {}
    for define in defs:
        key, _sep, value = define.partition('=')
        defines[key] = value
    return defines


def infixBinaryOp(handler, *args, **kwargs):
    """Handy wrapper to make sure binary operator handlers are called with only
    two arguments.

    Consecutive terms with the same operator are given as batch to the handler
    by pyparsing. E.g. 'a || b || c' will be seen as: [[a, '||', b, '||', c)]].
    This wrapper will recursively chop it up so that the given handler is
    always called with only two operands as expected for a binary operator.
    """

    def wrap(s, loc, toks):
        assert len(toks) == 1, toks
        toks = toks[0]
        while len(toks) > 3:
            toks = [ wrap(s, loc, [toks[0:3]]) ] + toks[3:]
        assert len(toks) == 3
        return handler(s, loc, toks, *args, **kwargs)

    return wrap

### Asyncio event loop setup

def dummy():
    pass

class EventLoopWrapper:
    def __init__(self):
        import asyncio
        import multiprocessing
        import signal
        import concurrent.futures

        if sys.platform == 'win32':
            loop = asyncio.ProactorEventLoop()
            asyncio.set_event_loop(loop)
            multiprocessing.set_start_method('spawn')
            executor = concurrent.futures.ProcessPoolExecutor()
        else:
            # The ProcessPoolExecutor is a barely usable for our interactive use
            # case. On SIGINT any busy executor should stop. The only way how this
            # does not explode is that we ignore SIGINT before spawning the process
            # pool and re-enable SIGINT in every executor. In the main process we
            # have to ignore BrokenProcessPool errors as we will likely hit them.
            # To "prime" the process pool a dummy workload must be executed because
            # the processes are spawned lazily.
            loop = asyncio.get_event_loop()
            origSigInt = signal.getsignal(signal.SIGINT)
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            # fork early before process gets big
            if sys.platform == 'msys':
                multiprocessing.set_start_method('fork')
            else:
                multiprocessing.set_start_method('forkserver')
            executor = concurrent.futures.ProcessPoolExecutor()
            executor.submit(dummy).result()
            signal.signal(signal.SIGINT, origSigInt)
        loop.set_default_executor(executor)

        self.__loop = loop
        self.__executor = executor

    def __enter__(self):
        return self.__loop

    def __exit__(self, exc_type, exc_value, traceback):
        self.__executor.shutdown()
        self.__loop.close()


async def run(args, universal_newlines=False, check=False, shell=False, **kwargs):
    """Provide the subprocess.run() function as asyncio corouting.

    This takes care of the missing 'universal_newlines' and 'check' options.
    Everything else is passed through. Will also raise the same exceptions as
    subprocess.run() to act as a drop-in replacement.
    """
    import asyncio
    import io
    import locale
    import subprocess

    if shell:
        proc = await asyncio.create_subprocess_shell(args, **kwargs)
    else:
        proc = await asyncio.create_subprocess_exec(*args, **kwargs)
    stdout, stderr = await proc.communicate()

    if universal_newlines and (stdout is not None):
        stdout = io.TextIOWrapper(io.BytesIO(stdout)).read()
    if universal_newlines and (stderr is not None):
        stderr = io.TextIOWrapper(io.BytesIO(stderr)).read()

    if check and (proc.returncode != 0):
        raise subprocess.CalledProcessError(proc.returncode, args,
            stdout, stderr)

    return subprocess.CompletedProcess(args, proc.returncode, stdout,
        stderr)

async def check_output(args, **kwargs):
    """The subprocess.check_output() call as coroutine."""
    import subprocess
    return (await run(args, check=True, stdout=subprocess.PIPE, **kwargs)).stdout



# Bob Build Tool
# Copyright (C) 2020  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

from .errors import BuildError
from .input import CheckoutAssert
from .scm import getScm
from .stringparser import Env
from .tty import Unbuffered
from .utils import removePath, emptyDirectory, isWindows
from enum import Enum
from pipes import quote
import asyncio
import concurrent.futures
import datetime
import io
import locale
import os
import shutil
import subprocess
import sys
import tempfile

__all__ = ['InvocationMode', 'Invoker']


class InvocationError(Exception):
    def __init__(self, returncode, what):
        self.returncode = returncode
        self.what = what

class CmdFailedError(InvocationError):
    def __init__(self, cmd, returncode):
        super().__init__(returncode, "Command '{}' returned exit status {}".format(cmd, returncode))
        self.cmd = cmd

class FinishedProcess:
    def __init__(self, returncode, stdout, stderr):
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr

class BlackHole:
    def write(self, data):
        pass
    def close(self):
        pass

DEVNULL = BlackHole()

class LogWriteProtocol(asyncio.SubprocessProtocol):
    def __init__(self, exitFuture, logFile, stdOut, stdErr):
        self.__exit = exitFuture
        self.__logFile = logFile
        self.__stdOut = stdOut
        self.__stdErr = stdErr

    def connection_made(self, transport):
        """Process is running and pipes have been connected"""
        self.__transport = transport

    def connection_lost(self, exc):
        """Process exited and all pipes have been disconnected"""
        self.__exit.set_result(self.__transport.get_returncode())
        self.__transport = None

    def pipe_data_received(self, fd, data):
        self.__logFile.write(data)
        if fd == 1:
            self.__stdOut.write(data)
        elif fd == 2:
            self.__stdErr.write(data)


class InvocationMode(Enum):
    CALL = 'call'
    SHELL = 'shell'

class Invoker:
    def __init__(self, spec, preserveEnv, noLogFiles, showStdOut, showStdErr, trace, redirect):
        self.__spec = spec
        self.__cwd = spec.workspaceWorkspacePath
        self.__preserveEnv = preserveEnv
        if preserveEnv:
            self.__env = os.environ.copy()
        else:
            self.__env = { k:v for (k,v) in os.environ.items()
                                         if k in spec.envWhiteList }
        self.__logFileName = None if noLogFiles else spec.logFile
        self.__logFile = DEVNULL
        self.__trace = trace
        self.__sandboxHelperPath = None
        self.__stdioBuffer = io.BytesIO() if redirect else None

        # Redirection is a bit complicated. We have to consider two levels: the
        # optional log file and the console.
        #
        #  * If there is a logfile then _all_ output is captured there.
        #  * What is printed is determined by the showStdOut/showStdErr knobs
        #  * Either the prints are sent to stdout/err or to the __stdioBuffer
        #    if all output is redirected/captured
        #  * If no log file and no redirection is used then the real tty should
        #    be used to let the called applications use all tty features (e.g.
        #    colors).
        if redirect:
            # Everything is redirected into a local buffer. Nothing is printed
            # on the console.
            self.__stdin = subprocess.DEVNULL
            if noLogFiles:
                self.__stdout = subprocess.PIPE if showStdOut else subprocess.DEVNULL
                self.__stderr = subprocess.PIPE if showStdErr else subprocess.DEVNULL
            else:
                self.__stdout = self.__stderr = subprocess.PIPE
            self.__stdoutStream = self.__stdioBuffer if showStdOut else DEVNULL
            self.__stderrStream = self.__stdioBuffer if showStdErr else DEVNULL
        else:
            # Depending on the verbosity some output is visible on the tty.
            self.__stdin = None
            if noLogFiles:
                self.__stdout = None if showStdOut else subprocess.DEVNULL
                self.__stderr = None if showStdErr else subprocess.DEVNULL
            else:
                self.__stdout = self.__stderr = subprocess.PIPE
            self.__stdoutStream = Unbuffered(sys.stdout.buffer) if showStdOut else DEVNULL
            self.__stderrStream = Unbuffered(sys.stderr.buffer) if showStdErr else DEVNULL

    def __openLog(self):
        # Create log file
        if self.__logFileName:
            self.__logFile = open(self.__logFileName, "ab", buffering=0)
            self.__logFile.write("### START: {}\n"
                .format(datetime.datetime.now().ctime())
                .encode(locale.getpreferredencoding()))
        else:
            self.__logFile = DEVNULL

    def __closeLog(self, lastExit):
        try:
            self.__logFile.write("### END({}): {}\n"
                .format(lastExit, datetime.datetime.now().ctime())
                .encode(locale.getpreferredencoding()))
            self.__logFile.close()
            self.__logFile = DEVNULL
        except OSError as e:
            self.error("Cannot close log file:", str(e))

    def __getSandboxHelperPath(self):
        if self.__sandboxHelperPath is None:
            # If Bob is run from the source directly we have to make sure that
            # the sandbox helper is up-to-date and use it from there. Otherwise
            # we assume that Bob is properly installed and that
            # bob-namespace-sandbox is in $PATH.
            try:
                from .develop.make import makeSandboxHelper
                sandboxHelper = makeSandboxHelper()
            except ImportError:
                # Determine absolute path here. We set $PATH when running in
                # the sandbox so we might not find it anymore.
                sandboxHelper = shutil.which("bob-namespace-sandbox")
                if sandboxHelper is None:
                    raise BuildError("Could not find bob-namespace-sandbox in $PATH! Please check your Bob installation.")

            self.__sandboxHelperPath = sandboxHelper

        return self.__sandboxHelperPath

    async def __runCommand(self, args, cwd, stdout=None, stderr=None,
                           check=False, env=None, universal_newlines=True,
                           **kwargs):
        cmd = " ".join(quote(a) for a in args)
        self.trace(cmd)
        if env is None:
            env = self.__env
        else:
            _env = self.__env.copy()
            _env.update(env)
            env = _env

        if stdout == True:
            # If stdout should be captured we use a dedicated buffer. This
            # buffer is then returned to the caller at child exit.
            stdoutRedir = subprocess.PIPE
            stdoutStream = io.BytesIO()
        elif stdout == False:
            stdoutRedir = subprocess.DEVNULL
            stdoutStream = DEVNULL
        else:
            stdoutRedir = self.__stdout
            stdoutStream = self.__stdoutStream

        if stderr == True:
            # If stderr should be captured we use a dedicated buffer. This
            # buffer is then returned to the caller at child exit.
            stderrRedir = subprocess.PIPE
            stderrStream = io.BytesIO()
        elif stdout == False:
            stderrRedir = subprocess.DEVNULL
            stderrStream = DEVNULL
        else:
            stderrRedir = self.__stderr
            stderrStream = self.__stderrStream

        # Sanity check on Windows that there are no environment variables that
        # differ only in case. The Windows envrionment is used case insensitive
        # by libc getenv() but the kernel passes the variables as-is.
        if isWindows():
            duplicates = []
            old = ""
            for i in sorted(env.keys(), key=str.upper):
                if i.upper() == old.upper():
                    duplicates.append((old, i))
                old = i
            if duplicates:
                self.fail("Colliding environment variables:",
                    ", ".join("{} ~= {}".format(i, j) for i,j in duplicates))

        loop = asyncio.get_event_loop()
        exitFuture = asyncio.Future()
        try:
            transport, protocol = await loop.subprocess_exec(
                lambda: LogWriteProtocol(exitFuture, self.__logFile,
                                         stdoutStream, stderrStream),
                *args,
                stdin=self.__stdin, stdout=stdoutRedir, stderr=stderrRedir,
                env=env, cwd=cwd,
                **kwargs)
        except OSError as e:
            self.fail(str(e), returncode=127)

        try:
            ret = await exitFuture
        finally:
            transport.close()
        if check and ret != 0:
            raise CmdFailedError(cmd, ret)

        if stdout == True:
            if universal_newlines:
                stdoutStream.seek(0)
                stdoutBuf = io.TextIOWrapper(stdoutStream).read()
            else:
                stdoutBuf = stdoutStream.getvalue()
        else:
            stdoutBuf = None

        if stderr == True:
            if universal_newlines:
                stderrStream.seek(0)
                stderrBuf = io.TextIOWrapper(stderrStream).read()
            else:
                stderrBuf = stderrStream.getvalue()
        else:
            stderrBuf = None

        return FinishedProcess(ret, stdoutBuf, stderrBuf)

    def __getSandboxCmds(self, tmpDir):
        if sys.platform != "linux":
            self.fail("Sandbox builds are only supported on Linux!")

        cmdArgs = [ self.__getSandboxHelperPath() ]
        #FIXME: if verbosity >= 4: cmdArgs.append('-D')
        cmdArgs.extend(["-S", tmpDir])
        cmdArgs.extend(["-H", "bob"])
        cmdArgs.extend(["-d", "/tmp"])
        sandboxRootFs = os.path.abspath(self.__spec.sandboxRootWorkspace)
        for f in os.listdir(sandboxRootFs):
            cmdArgs.extend(["-M", os.path.join(sandboxRootFs, f), "-m", "/"+f])

        skipOpt = "nojenkins" if self.__spec.isJenkins else "nolocal"
        substEnv = Env(self.__env)
        for (hostPath, sndbxPath, options) in self.__spec.sandboxHostMounts:
            if skipOpt in options: continue
            hostPath = substEnv.substitute(hostPath, hostPath, False)
            if "nofail" in options:
                if not os.path.exists(hostPath): continue
            sndbxPath = substEnv.substitute(sndbxPath, sndbxPath, False)
            cmdArgs.extend(["-M", hostPath])
            if "rw" in options:
                cmdArgs.extend(["-w", sndbxPath])
            elif hostPath != sndbxPath:
                cmdArgs.extend(["-m", sndbxPath])

        return cmdArgs

    async def executeStep(self, mode, clean=False, keepSandbox=False):
        # make permissions predictable
        os.umask(0o022)

        tmpDir = None
        ret = 1
        try:
            self.__openLog()

            # Create temporary directory for the script file (or others as
            # needed by the particular scripting engine). This directory is
            # also used as ephemeral sandbox container.
            tmpDir = tempfile.mkdtemp()

            # prepare workspace
            clean = self.__spec.clean if self.__spec.clean is not None else clean
            if not os.path.isdir(self.__spec.workspaceWorkspacePath):
                os.makedirs(self.__spec.workspaceWorkspacePath, exist_ok=True)
            elif clean and mode != InvocationMode.SHELL:
                emptyDirectory(self.__spec.workspaceWorkspacePath)

            # setup script and arguments
            if mode == InvocationMode.SHELL:
                realScriptFile, execScriptFile, callArgs = self.__spec.language.setupShell(
                    self.__spec, tmpDir, self.__preserveEnv)
            elif mode == InvocationMode.CALL:
                realScriptFile, execScriptFile, callArgs = self.__spec.language.setupCall(
                    self.__spec, tmpDir, self.__preserveEnv, self.__trace)
            else:
                assert False, "not reached"

            # Wrap call into sandbox if requested
            if self.__spec.hasSandbox:
                cmdArgs = self.__getSandboxCmds(tmpDir)
                cmdArgs.extend(["-M", os.path.abspath(realScriptFile), "-m"
                    , execScriptFile])

                # Prevent network access
                if not self.__spec.sandboxNetAccess: cmdArgs.append('-n')

                # Create empty env file. Otheriwse bind mount fails.
                with open(self.__spec.envFile, "wb"): pass
                cmdArgs.extend(["-M", os.path.abspath(self.__spec.envFile), "-w", "/bob/env"])

                # Mount workspace writable and all dependencies read-only
                cmdArgs.extend(["-M", os.path.abspath(self.__spec.workspaceWorkspacePath),
                    "-w", self.__spec.workspaceExecPath])
                cmdArgs.extend(["-W", os.path.abspath(self.__spec.workspaceExecPath) ])
                for argWorkspacePath, argExecPath in self.__spec.sandboxDepMounts:
                    cmdArgs.extend(["-M", os.path.abspath(argWorkspacePath),
                        "-m", argExecPath])

                # Command follows. Stop parsing options.
                cmdArgs.append("--")
            else:
                cmdArgs = []
            cmdArgs.extend(callArgs)

            if mode == InvocationMode.SHELL:
                ret = await self.callCommand(cmdArgs)
            elif mode == InvocationMode.CALL:
                for scm in self.__spec.preRunCmds:
                    scm = getScm(scm)
                    try:
                        await scm.invoke(self)
                    except CmdFailedError as e:
                        self.error(scm.getSource(), "failed")
                        self.error(e.what)
                        raise
                    except Exception:
                        self.error(scm.getSource(), "failed")
                        raise
                await self.checkCommand(cmdArgs)
                for a in self.__spec.postRunCmds:
                    a = CheckoutAssert(a)
                    try:
                        await a.invoke(self)
                    except Exception:
                        self.error(a.getSource(), "failed")
                        raise
            else:
                assert False, "not reached"

            # everything went well
            ret = 0

        except OSError as e:
            self.error("Something went wrong:", str(e))
            ret = 1
        except InvocationError as e:
            ret = e.returncode
        finally:
            if tmpDir is not None:
                if not self.__spec.hasSandbox or not keepSandbox:
                    try:
                        removePath(tmpDir)
                    except OSError as e:
                        self.error("Error removing sandbox:", str(e))
                elif self.__spec.hasSandbox:
                    self.info("Keeping sandbox image at", tmpDir)
            self.__closeLog(ret)

        return ret

    async def executeFingerprint(self, keepSandbox=False):
        # make permissions predictable
        os.umask(0o022)

        tmpDir = None
        proc = None
        ret = -1
        stdout = stderr = b''
        try:
            self.__openLog()

            # The fingerprint is always exectuted in a temporary directory
            tmpDir = tempfile.mkdtemp(dir=os.getcwd(), prefix=".bob-")

            # Wrap call into sandbox if requested
            env = {}
            if self.__spec.hasSandbox:
                env["BOB_CWD"] = "/bob/fingerprint"
                env["PATH"] = ":".join(self.__spec.sandboxPaths)

                # Setup workspace
                cmdArgs = self.__getSandboxCmds(tmpDir)
                cmdArgs.extend(["-d", "/bob/fingerprint"])
                cmdArgs.extend(["-W", "/bob/fingerprint"])

                # Prevent network access
                cmdArgs.append('-n')

                # Command follows. Stop parsing options.
                cmdArgs.append("--")
            else:
                env["BOB_CWD"] = tmpDir
                cmdArgs = []

            cmdArgs += self.__spec.language.setupFingerprint(self.__spec, env)
            proc = await self.__runCommand(cmdArgs, tmpDir, env=env,
                universal_newlines=False, stdout=True, stderr=True)
            ret = proc.returncode
            stdout = proc.stdout
            stderr = proc.stderr

        except InvocationError as e:
            ret = e.returncode
            raise BuildError(e.what)
        finally:
            if tmpDir is not None:
                if not self.__spec.hasSandbox or not keepSandbox:
                    try:
                        removePath(tmpDir)
                    except OSError as e:
                        self.error("Error removing sandbox:", str(e))
                elif self.__spec.hasSandbox:
                    self.info("Keeping sandbox image at", tmpDir)
            self.__closeLog(ret)

        return (ret, stdout, stderr)

    async def callCommand(self, args, cwd=None, **kwargs):
        cwd = os.path.join(self.__cwd, cwd) if cwd else self.__cwd
        ret = await self.__runCommand(args, cwd, **kwargs)
        return ret.returncode

    async def checkCommand(self, args, cwd=None, **kwargs):
        cwd = os.path.join(self.__cwd, cwd) if cwd else self.__cwd
        await self.__runCommand(args, cwd, check=True, **kwargs)

    async def checkOutputCommand(self, args, cwd=None, **kwargs):
        cwd = os.path.join(self.__cwd, cwd) if cwd else self.__cwd
        ret = await self.__runCommand(args, cwd, stdout=True, check=True, **kwargs)
        return ret.stdout.rstrip()

    def __print(self, *args, file=DEVNULL, **kwargs):
        buf = io.StringIO()
        print(*args, **kwargs, file=buf)
        binbuf = buf.getvalue().encode(locale.getpreferredencoding())
        file.write(binbuf)
        self.__logFile.write(binbuf)
        return buf.getvalue()

    def trace(self, *args, **kwargs):
        if self.__trace:
            return self.__print("+", *args, **kwargs, file=self.__stderrStream)

    def info(self, *args, **kwargs):
        return self.__print("Info:", *args, **kwargs, file=self.__stderrStream)

    def warn(self, *args, **kwargs):
        return self.__print("Warning:", *args, **kwargs, file=self.__stderrStream)

    def error(self, *args, **kwargs):
        return self.__print("Error:", *args, **kwargs, file=self.__stderrStream)

    def fail(self, *args, returncode=1, **kwargs):
        what = self.error(*args, **kwargs)
        raise InvocationError(returncode, what)

    def joinPath(self, *paths):
        return os.path.join(self.__cwd, *paths)

    def getStdio(self):
        return self.__stdioBuffer.getvalue().decode(
            locale.getpreferredencoding(), 'surrogateescape')




# Bob build tool
# Copyright (C) 2016  TechniSat Digital GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later

from .tty import colorize

class BobError(Exception):
    def __init__(self, slogan, kind=None, stackSlogan="", help="", returncode=1):
        self.kind = (kind + " error: ") if kind is not None else "Error: "
        self.slogan = slogan
        self.stackSlogan = stackSlogan
        self.stack = []
        self.help = help
        self.returncode = returncode

    def __str__(self):
        ret = colorize(self.kind, "31;1") + colorize(self.slogan, "31")
        if self.stack:
            ret = ret + "\n" + self.stackSlogan + ": " + "/".join(self.stack)
        if self.help:
            ret = ret + "\n" + self.help
        return ret

    def pushFrame(self, frame):
        if not self.stack or (self.stack[0] != frame):
            self.stack.insert(0, frame)

    def setStack(self, stack):
        if not self.stack: self.stack = stack[:]

class ParseError(BobError):
    def __init__(self, slogan, *args, **kwargs):
        BobError.__init__(self, slogan, "Parse", "Processing stack", *args, **kwargs)

class BuildError(BobError):
    def __init__(self, slogan, *args, **kwargs):
        BobError.__init__(self, slogan, "Build", "Failed package", *args, **kwargs)


class MultiBobError(BobError):
    def __init__(self, others):
        self.others = []
        for i in others:
            if isinstance(i, MultiBobError):
                for j in i.others:
                    if j not in self.others: self.others.append(j)
            elif isinstance(i, BobError):
                if i not in self.others: self.others.append(i)
            else:
                raise i

    def __str__(self):
        return "\n".join(str(i) for i in self.others)

    def pushFrame(self, frame):
        pass

    def setStack(self, stack):
        pass

    @property
    def returncode(self):
        return max(1, 1, *(i.returncode for i in self.others))

